-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THMoistureSystem = {}
local THMoistureSystem_mt = THUtils.createClass(THMoistureSystem, THAbstractSystem)
local debugFlagId = THDebugUtil.createDebugFlag("THMoistureSystem")
THMoistureSystem.debugFlagId = debugFlagId
THMoistureSystem.DATA_KEY = "moistureSystem"
THMoistureSystem.FACTOR_SEND_BITS = 16
THMoistureSystem.PERIOD_SEND_BITS = 6
THMoistureSystem.DRYING_STATE_BITS = 4
THMoistureSystem.DEFAULT_DRYING_FACTORS = {
    NIGHT = 0.5,
    CLOUD = 0.5,
    RAIN  = 0,
    STORM = 0,
    SNOW = 0.25,
    WIND = 1
}
THMoistureSystem.SETTING = {
    PRICE_BONUS_FACTOR = "priceBonusFactor",
    PRICE_PENALTY_FACTOR = "pricePenaltyFactor",
    YIELD_BONUS_FACTOR = "yieldBonusFactor",
    YIELD_PENALTY_FACTOR = "yieldPenaltyFactor",
    AUTO_STOP_DRYERS = "autoStopDryers",
    STORAGE_DRYING_FACTOR = "storageDryingFactor",
    STORAGE_LOSS_FACTOR = "storageLossFactor",
    STORAGE_LOSS_FACTOR_DRYING = "storageLossFactorDrying"
}
function THMoistureSystem.initXMLSchema(xmlSchema, xmlPath)
    THMoistureSystem:superClass().initXMLSchema(xmlSchema, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath .. ".environment", "#minHumidity", "Absolute minimum allowed humidity percentage", 30)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath .. ".environment", "#dryingSpeed", "How fast environment dries after rain", 1)
    local function registerFruitTypeDryingStates(pBasePath)
        local dryingStatePath = pBasePath .. ".dryingStates.state(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, dryingStatePath, "#min", "Minimum fruitType growth state")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, dryingStatePath, "#max", "Maximum fruitType growth state")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#pctStart", "FruitType starting moisture percentage", 100)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#pctEnd", "FruitType ending moisture percentage", 0)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#hoursToDry", "Hours for crop to dry under optimal conditions")
        local resetPeriodPath = dryingStatePath .. ".resetPeriods.period(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, resetPeriodPath, "#id", "Game period id/index")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, resetPeriodPath, "reset", "Drying state reset flag", true)
        local dryingFactorsPath = dryingStatePath .. ".baseFactors"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#night", "Amount that night affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.NIGHT)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#cloud", "Amount that clouds affect dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.CLOUD)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#rain", "Amount that rain affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.RAIN)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#storm", "Amount that hail/storms affect dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.STORM)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#snow", "Amount that snow affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.SNOW)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingFactorsPath, "#wind", "Amount that wind affects dry down", THMoistureSystem.DEFAULT_DRYING_FACTORS.WIND)
        local humidityFactorPath = dryingStatePath .. ".humidityFactors.humidity(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, humidityFactorPath, "#value", "Humidity percentage")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, humidityFactorPath, "#factor", "Drying factor at this humidity percentage")
        local soilTypeFactorPath = dryingStatePath .. ".soilTypeFactors.soilType(?)"
        THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, soilTypeFactorPath, "#index", "SoilType index")
        THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, soilTypeFactorPath, "factor", "SoilType factor")
    end
    local fruitTypePath = xmlPath .. ".fruitTypes.fruitType(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fruitTypePath, "#name", "FruitType name")
    registerFruitTypeDryingStates(fruitTypePath)
    local fillTypePath = xmlPath .. ".fillTypes.fillType(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fillTypePath, "#name", "FillType name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypePath, "#min", "Mininum moisture percentage")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypePath, "#max", "Maximum moisture percentage")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypePath, "#default", "Default moisture percentage")
    local fillTypeFactorPath = fillTypePath .. ".factors.factor(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#moisture", "FillType moisture percentage")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#price", "Price factor at specified moisture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#yield", "Yield factor at specified moisture")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeFactorPath, "#decay", "Storage loss per in-game period at specified moisture")
end
function THMoistureSystem.initXMLSchemaSavegame(xmlSchema, xmlPath)
    THMoistureSystem:superClass().initXMLSchemaSavegame(xmlSchema, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#cachedDryingTime", "Cached drying time", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, xmlPath, "#periodLength", "Last daysPerPeriod setting", 1)
    local dryingConditionsPath = xmlPath .. ".dryingConditions"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, dryingConditionsPath, "#isRaining", "Last isRaining value", false)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, dryingConditionsPath, "#isDayTime", "Last isSunOn value", false)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingConditionsPath, "#temp", "Last temperature value", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingConditionsPath, "#dewPoint", "Last dew point value", 0)
    local farmlandFactorPath = dryingConditionsPath .. ".farmlandFactors.farmland(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, farmlandFactorPath, "#id", "Farmland id")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, farmlandFactorPath, "#factor", "Farmland drying factor")
    local currentRainPath = dryingConditionsPath .. ".currentRain"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#startDay", "Game day when rain event started", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#startTime", "Game time of day when rain event started", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#endDay", "Game day when rain event ended", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#endTime", "Game time of day when rain event ended", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, currentRainPath, "#duration", "Total game time that rain event lasted", 0)
    local lastRainPath = dryingConditionsPath .. ".lastRain"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#startDay", "Game day when rain event started", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#startTime", "Game time of day when rain event started", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#endDay", "Game day when rain event ended", 1)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#endTime", "Game time of day when rain event ended", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, lastRainPath, "#duration", "Total game time that rain event lasted", 0)
    local fruitTypePath = xmlPath .. ".fruitTypes.fruitType(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fruitTypePath, "#name", "FruitType name")
    local dryingStatePath = fruitTypePath .. ".dryingStates.dryingState(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.INT, dryingStatePath, "#index", "Drying state index")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#averageTemp", "Average temperature", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingStatePath, "#averageWind", "Average wind speed", 0)
    local dryingTimesPath = dryingStatePath .. ".dryingTimes"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#total", "Total drying hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#night", "Total night hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#cloud", "Total cloud hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#rain", "Total rain hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#storm", "Total storm hours", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryingTimesPath, "#snow", "Total snow hours", 0)
end
function THMoistureSystem.new(rootEnv, dataKey, modName, customMt)
    customMt = customMt or THMoistureSystem_mt
    local self = THAbstractSystem.new(rootEnv, dataKey, modName, customMt)
    self.thFruitTypeManager = rootEnv.thFruitTypeManager
    self.thFillTypeManager = rootEnv.thFillTypeManager
    self.thPFManager = rootEnv.thPFManager
    self.dryingConditions = {
        isDayTime = false,
        isRaining = false,
        humidity = 0,
        dewPoint = 0,
        temperature = 0,
        absMinHumidity = 0.3,
        absMaxHumidity = 1,
        maxDryingTime = 720, -- 12 in-game hours before/after a rain
        dryingTimeFactor = 1,
        dryingPeriods = {},
        farmlandFactors = {},
        isInitialized = false
    }
    self.dryingConditions.lastRainData = {
        startTime = 0,
        endTime = 0,
        startDay = 1,
        endDay = 1,
        duration = 0
    }
    self.dryingConditions.currentRainData = {
        startTime = 0,
        endTime = 0,
        startDay = 1,
        endDay = 1,
        duration = 0
    }
    self.fruitTypes = {}
    self.dryingStateCachedTime = 0
    self.isDryingStateUpdateSafe = false
    self.areDryingStatesInitialized = false
    self.periodLength = 1
    self.resetPeriodIndexToId = {}
    for periodId, periodIndex in pairs(Environment.PERIOD) do
        self.resetPeriodIndexToId[periodIndex] = periodId
    end
    self.fillTypes = {}
    self.moneyChangeTypes = {
        harvestMoisture = MoneyType.register("harvestIncome", "thFinance_harvestMoisture", modName),
        dryingCosts = MoneyType.register("harvestIncome", "thFinance_dryingCosts", modName)
    }
    self.dirtyFlag = self:getNextDirtyFlag()
    self.messageCenter:subscribe(MessageType.MINUTE_CHANGED, self.onMinuteChanged, self)
    self.messageCenter:subscribe(MessageType.HOUR_CHANGED, self.onHourChanged, self)
    self.messageCenter:subscribe(MessageType.DAY_CHANGED, self.onDayChanged, self)
    self.messageCenter:subscribe(MessageType.PERIOD_CHANGED, self.onPeriodChanged, self)
    self.messageCenter:subscribe(MessageType.WEATHER_CHANGED, self.onWeatherChanged, self)
    return self
end
function THMoistureSystem.newModule(rootEnv, modData)
    if _G.g_thMoistureSystem ~= nil then
        THUtils.warningMsg("Module %q already loaded", THMoistureSystem.DATA_KEY)
    else
        local self = THMoistureSystem.new(rootEnv, THMoistureSystem.DATA_KEY, modData.name)
        if self ~= nil and self:addSchemaFunctions() then
            _G.g_thMoistureSystem = self
            local modName = self.modName
            local modPath = self.modPath
            source(modPath .. "scripts/gui/THInGameMoistureWeatherFrame.lua")
            source(modPath .. "scripts/gui/THFillLevelMoistureDisplay.lua")
            source(modPath .. "scripts/gui/THPlayerMoistureHud.lua")
            source(modPath .. "scripts/objects/THMoistureStorage.lua")
            source(modPath .. "scripts/objects/THMoistureUnloadingStation.lua")
            source(modPath .. "scripts/objects/THMoistureLoadingStation.lua")
            source(modPath .. "scripts/objects/THStorageDryer.lua")
            local thSpecManager = self.thSpecManager
            thSpecManager:addSpecialization("vehicle", "moistureFillUnit", "THMoistureFillUnit", modName, "scripts/vehicles/specializations/THMoistureFillUnit.lua")
            thSpecManager:addSpecialization("vehicle", "moistureDischargeable", "THMoistureDischargeable", modName, "scripts/vehicles/specializations/THMoistureDischargeable.lua")
            thSpecManager:addSpecialization("vehicle", "moistureCombine", "THMoistureCombine", modName, "scripts/vehicles/specializations/THMoistureCombine.lua")
            thSpecManager:addSpecialization("placeable", "moistureStorage", "THPlaceableMoistureStorage", modName, "scripts/placeables/specializations/THPlaceableMoistureStorage.lua")
            return self
        end
    end
end
function THMoistureSystem:initializeDataTables()
    if self:initFillTypeData() and self:initFruitTypeData() then
        return true
    end
    return false
end
function THMoistureSystem:initFruitTypeData()
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypesArray, numFruitTypes = thFruitTypeManager:getFruitTypeList(true)
    if fruitTypesArray ~= nil and numFruitTypes > 0 then
        for fruitTypeIndex = 1, numFruitTypes do
            local fruitTypeInfo = fruitTypesArray[fruitTypeIndex]
            local fruitTypeData = thFruitTypeManager:getFruitTypeData(fruitTypeIndex)
            local fruitTypeMoistureData = {
                dryingStates = {},
                dryingStateMapping = {},
                isEnabled = true
            }
            fruitTypeMoistureData.fruitType = {
                name = fruitTypeInfo.name,
                index = fruitTypeInfo.index,
                info = fruitTypeInfo,
                data = fruitTypeData
            }
            if fruitTypeData == nil then
                fruitTypeMoistureData.isEnabled = false
            end
            self.fruitTypes[fruitTypeIndex] = fruitTypeMoistureData
        end
    end
    return true
end
function THMoistureSystem:initFillTypeData()
    local thFillTypeManager = self.thFillTypeManager
    local fillTypesArray, numFillTypes = thFillTypeManager:getFillTypes()
    if fillTypesArray ~= nil and numFillTypes > 0 then
        for fillTypeIndex = 1, numFillTypes do
            local fillTypeInfo = fillTypesArray[fillTypeIndex]
            local fillTypeData = thFillTypeManager:getFillTypeData(fillTypeIndex)
            local fillTypeMoistureData = {
                minPercent = 0,
                maxPercent = 1,
                defPercent = 0.5,
                hasMoisture = false,
                isEnabled = true
            }
            fillTypeMoistureData.fillType = {
                name = fillTypeInfo.name,
                title = fillTypeInfo.title,
                index = fillTypeInfo.index,
                info = fillTypeInfo,
                data = fillTypeData
            }
            fillTypeMoistureData.factors = {
                curve = {},
                minPrice = 1,
                maxPrice = 1,
                minYield = 1,
                maxYield = 1,
                minDecay = 0,
                maxDecay = 0
            }
            if fillTypeData == nil then
                fillTypeMoistureData.isEnabled = false
            end
            self.fillTypes[fillTypeIndex] = fillTypeMoistureData
        end
    end
    return true
end
function THMoistureSystem:finalizeDataTables()
    if self:finalizeFruitTypeData() and self:finalizeFillTypeData() then
        if THDebugUtil.getIsEnabled(debugFlagId) then
            THUtils.displayMsg("Moisture System Environment")
            THUtils.printTable(self)
            THUtils.displayMsg("- Drying Conditions:")
            THUtils.printTable(self.dryingConditions, 1)
            for _, fruitTypeMoistureData in ipairs(self.fruitTypes) do
                local fruitTypeInfo = fruitTypeMoistureData.fruitType.info
                THUtils.displayMsg("Moisture values for fruitType %q [%s]:", fruitTypeInfo.name, fruitTypeInfo.index)
                THUtils.printTable(fruitTypeMoistureData)
                THUtils.displayMsg("- Drying States:")
                for _, dryingStateData in ipairs(fruitTypeMoistureData.dryingStates) do
                    THUtils.printTable(dryingStateData)
                    THUtils.displayMsg("- Drying Times:")
                    THUtils.printTable(dryingStateData.dryingTimes, 1)
                    THUtils.displayMsg("- Drying Factors:")
                    THUtils.printTable(dryingStateData.dryingFactors, 4)
                    THUtils.displayMsg("- Reset Periods:")
                    THUtils.printTable(dryingStateData.resetPeriods, 4)
                end
            end
            for _, fillTypeMoistureData in ipairs(self.fillTypes) do
                local fillTypeInfo = fillTypeMoistureData.fillType.info
                THUtils.displayMsg("Moisture values for fillType %q [%s]:", fillTypeInfo.name, fillTypeInfo.index)
                THUtils.printTable(fillTypeMoistureData)
                THUtils.displayMsg("- Moisture factors:")
                THUtils.printTable(fillTypeMoistureData.factors, 4)
            end
        end
        return true
    end
    return false
end
function THMoistureSystem:finalizeFruitTypeData()
    local thFruitTypeManager = self.thFruitTypeManager
    local growthSystem = self.growthSystem
    local maxNumPeriods = Environment.PERIODS_IN_YEAR
    local autoConfiguredCrops = {}
    for fruitTypeIndex, fruitTypeMoistureData in ipairs(self.fruitTypes) do
        local fruitTypeData = fruitTypeMoistureData.fruitType.data
        local fillTypeInfo = thFruitTypeManager:getFruitTypeFillType(fruitTypeIndex)
        local fillTypeMoistureData = self:getFillTypeMoistureData(fillTypeInfo, true) -- force
        THUtils.clearTable(fruitTypeMoistureData.dryingStateMapping) -- safety
        if fruitTypeData == nil or fillTypeInfo == nil or fillTypeMoistureData == nil then
            THUtils.clearTable(fruitTypeMoistureData.dryingStates)
            fruitTypeMoistureData.isEnabled = false
        else
            local minHarvestState = fruitTypeData.minHarvestState or -1
            local maxHarvestState = fruitTypeData.maxHarvestState or -1
            local hasDryingStates = #fruitTypeMoistureData.dryingStates > 0
            if minHarvestState < 0 or maxHarvestState < 0 then
                THUtils.clearTable(fruitTypeMoistureData.dryingStates)
                hasDryingStates = false
            elseif not hasDryingStates then
                local dryingStateData = self:createFruitTypeDryingState(fruitTypeIndex, minHarvestState, maxHarvestState, 0.3, 0.1)
                if dryingStateData ~= nil then
                    dryingStateData.index = #fruitTypeMoistureData.dryingStates + 1
                    fruitTypeMoistureData.dryingStates[dryingStateData.index] = dryingStateData
                    autoConfiguredCrops[fruitTypeMoistureData] = true
                    hasDryingStates = true
                end
            end
            if hasDryingStates then
                local minHarvestMoisture = math.huge
                local maxHarvestMoisture = -math.huge
                for dryingStateIndex, dryingStateData in ipairs(fruitTypeMoistureData.dryingStates) do
                    local minGrowthState = dryingStateData.minGrowthState
                    local maxGrowthState = dryingStateData.maxGrowthState
                    local resetPeriods = dryingStateData.resetPeriods
                    dryingStateData.index = dryingStateIndex
                    if minGrowthState >= minHarvestState or maxGrowthState <= maxHarvestState then
                        minHarvestMoisture = math.min(minHarvestMoisture, dryingStateData.startPercent, dryingStateData.endPercent)
                        maxHarvestMoisture = math.max(maxHarvestMoisture, dryingStateData.startPercent, dryingStateData.endPercent)
                    end
                    for growthState = minGrowthState, maxGrowthState do
                        fruitTypeMoistureData.dryingStateMapping[growthState] = dryingStateIndex
                    end
                    if maxNumPeriods == 0 then
                        resetPeriods = THUtils.clearTable(dryingStateData.resetPeriods)
                        dryingStateData.hasResetPeriods = false
                    elseif dryingStateData.hasResetPeriods and next(resetPeriods) == nil then
                        local growthData = growthSystem.seasonalFruitData[fruitTypeIndex]
                        dryingStateData.hasResetPeriods = false -- reset
                        if growthData ~= nil then
                            for period = 1, maxNumPeriods do
                                local periodData = growthData[period]
                                if periodData ~= nil and periodData.harvestPossible then
                                    local lastPeriodData = nil -- placeholder
                                    dryingStateData.hasResetPeriods = true
                                    if period == 1 then
                                        lastPeriodData = growthData[maxNumPeriods]
                                    else
                                        lastPeriodData = growthData[period - 1]
                                    end
                                    if lastPeriodData == nil or not lastPeriodData.harvestPossible then
                                        resetPeriods[period] = true
                                    end
                                end
                            end
                            if dryingStateData.hasResetPeriods then
                                if next(resetPeriods) == nil then
                                    for period = 1, maxNumPeriods do
                                        resetPeriods[period] = true
                                    end
                                end
                                autoConfiguredCrops[fruitTypeMoistureData] = true
                            end
                        end
                    end
                end
                if fillTypeMoistureData.isEnabled
                    and not fillTypeMoistureData.hasMoisture
                    and minHarvestMoisture < math.huge and maxHarvestMoisture > -math.huge
                then
                    THUtils.clearTable(fillTypeMoistureData.factors.curve)
                    if self:setFillTypeMoistureFactors(fillTypeInfo, minHarvestMoisture, 1, 1, 0)
                        and self:setFillTypeMoistureFactors(fillTypeInfo, maxHarvestMoisture, 0.8, 0.8, 0.05)
                    then
                        fillTypeMoistureData.minPercent = minHarvestMoisture
                        fillTypeMoistureData.maxPercent = maxHarvestMoisture
                        fillTypeMoistureData.defPercent = (minHarvestMoisture + maxHarvestMoisture) / 2
                        fillTypeMoistureData.hasMoisture = true
                        autoConfiguredCrops[fruitTypeMoistureData] = true
                    end
                end
            end
        end
    end
    for fruitTypeMoistureData in pairs(autoConfiguredCrops) do
        THUtils.displayMsg("Moisture values auto-generated for fruitType %q [%s]:", fruitTypeMoistureData.fruitType.name, fruitTypeMoistureData.fruitType.index)
    end
    return true
end
function THMoistureSystem:finalizeFillTypeData()
    for _, fillTypeMoistureData in ipairs(self.fillTypes) do
        local moistureFactors = fillTypeMoistureData.factors
        local numCurveEntries = #moistureFactors.curve
        if numCurveEntries > 0 then
            local minPrice = math.huge
            local minYield = math.huge
            local minDecay = math.huge
            local maxPrice = -math.huge
            local maxYield = -math.huge
            local maxDecay = -math.huge
            for curveIndex = 1, numCurveEntries do
                local curveEntry = moistureFactors.curve[curveIndex]
                local priceFactor = THUtils.getNoNil(curveEntry.priceFactor, 1)
                local yieldFactor = THUtils.getNoNil(curveEntry.yieldFactor, 1)
                local decayFactor = THUtils.getNoNil(curveEntry.decayFactor, 0)
                priceFactor = math.max(0, priceFactor)
                yieldFactor = math.max(0, yieldFactor)
                decayFactor = math.max(0, decayFactor)
                curveEntry.priceFactor = priceFactor
                curveEntry.yieldFactor = yieldFactor
                curveEntry.decayFactor = decayFactor
                minPrice = math.min(minPrice, priceFactor)
                minYield = math.min(minYield, yieldFactor)
                minDecay = math.min(minDecay, decayFactor)
                maxPrice = math.max(maxPrice, priceFactor)
                maxYield = math.max(maxYield, yieldFactor)
                maxDecay = math.max(maxDecay, decayFactor)
            end
            moistureFactors.minPrice = math.min(minPrice, maxPrice)
            moistureFactors.minYield = math.min(minYield, maxYield)
            moistureFactors.minDecay = math.min(minDecay, maxDecay)
            moistureFactors.maxPrice = math.max(minPrice, maxPrice)
            moistureFactors.maxYield = math.max(minYield, maxYield)
            moistureFactors.maxDecay = math.max(minDecay, maxDecay)
        end
    end
    return true
end
function THMoistureSystem:initSettings()
    local thSettingsManager = self.thSettingsManager
    local categoryMain = thSettingsManager:addSettingCategory(self.dataKey, nil, self.modName)
    local settingPriceBonusFactor = thSettingsManager:addSetting(THMoistureSystem.SETTING.PRICE_BONUS_FACTOR, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    local settingPricePenaltyFactor = thSettingsManager:addSetting(THMoistureSystem.SETTING.PRICE_PENALTY_FACTOR, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    settingPriceBonusFactor:addValue(0, "$l10n_ui_off")
    settingPricePenaltyFactor:addValue(0, "$l10n_ui_off")
    local settingYieldBonusFactor = thSettingsManager:addSetting(THMoistureSystem.SETTING.YIELD_BONUS_FACTOR, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    local settingYieldPenaltyFactor = thSettingsManager:addSetting(THMoistureSystem.SETTING.YIELD_PENALTY_FACTOR, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    settingYieldBonusFactor:addValue(0, "$l10n_ui_off")
    settingYieldPenaltyFactor:addValue(0, "$l10n_ui_off")
    local settingAutoStopDryers = thSettingsManager:addSetting(THMoistureSystem.SETTING.AUTO_STOP_DRYERS, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    settingAutoStopDryers:addValue(THStorageDryer.AUTO_STOP.NEVER, "$l10n_thUI_never")
    settingAutoStopDryers:addValue(THStorageDryer.AUTO_STOP.DEFAULT, "$l10n_thUI_default")
    settingAutoStopDryers:addValue(THStorageDryer.AUTO_STOP.ALWAYS, "$l10n_thUI_always")
    settingAutoStopDryers:setValueByIndex(2, true, true) -- default
    local settingStorageDryingFactor = thSettingsManager:addSetting(THMoistureSystem.SETTING.STORAGE_DRYING_FACTOR, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    local settingStorageLossFactor = thSettingsManager:addSetting(THMoistureSystem.SETTING.STORAGE_LOSS_FACTOR, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    local settingStorageLossDryingFactor = thSettingsManager:addSetting(THMoistureSystem.SETTING.STORAGE_LOSS_FACTOR_DRYING, categoryMain, nil, nil, THSettingsManager.ELEMENT_TYPE.MULTI_OPTION, self.modName, false, true, true)
    settingStorageLossFactor:addValue(0, "$l10n_ui_off")
    settingStorageLossDryingFactor:addValue(0, "$l10n_ui_off")
    for i = 1, 20 do -- 0.25 to 5.0
        local factorValue = THUtils.round(i * 0.25, 2)
        local factorText = THUtils.formatPercent(factorValue)
        if factorText == "100%" then
            factorText = "$l10n_thUI_default"
        end
        settingPriceBonusFactor:addValue(factorValue, factorText)
        settingPricePenaltyFactor:addValue(factorValue, factorText)
        settingYieldBonusFactor:addValue(factorValue, factorText)
        settingYieldPenaltyFactor:addValue(factorValue, factorText)
        settingStorageDryingFactor:addValue(factorValue, factorText)
        settingStorageLossFactor:addValue(factorValue, factorText)
        settingStorageLossDryingFactor:addValue(factorValue, factorText)
    end
    settingPriceBonusFactor:setValue(1, true, true)
    settingPricePenaltyFactor:setValue(1, true, true)
    settingYieldBonusFactor:setValue(1, true, true)
    settingYieldPenaltyFactor:setValue(1, true, true)
    settingStorageDryingFactor:setValue(1, true, true)
    settingStorageLossFactor:setValue(1, true, true)
    settingStorageLossDryingFactor:setValue(1, true, true)
end
function THMoistureSystem:loadFromConfigXMLFile(xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local thFillTypeManager = self.thFillTypeManager
    local loadSuccess = THMoistureSystem:superClass().loadFromConfigXMLFile(self, xmlFile, xmlKey)
    if loadSuccess then
        local dryingConditions = self.dryingConditions
        local environmentKey = xmlKey .. ".environment"
        local absMinHumidity = THUtils.getXMLValue(xmlFile, environmentKey, "#minHumidity", 30)
        local dryingTimeFactor = THUtils.getXMLValue(xmlFile, environmentKey, "#dryingSpeed", 1)
        if absMinHumidity < 0 or absMinHumidity > 100 then
            THUtils.xmlMsg(environmentKey, "minHumidity", absMinHumidity, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
        else
            dryingConditions.absMinHumidity = MathUtil.clamp(absMinHumidity / 100, 0, 1)
        end
        if dryingTimeFactor <= 0 then
            THUtils.xmlMsg(environmentKey, "dryingSpeed", dryingTimeFactor, true, THXMLMessage.VALUE_GREATER_ZERO)
        else
            dryingConditions.dryingTimeFactor = dryingTimeFactor
        end
    end
    if loadSuccess then
        xmlFile:iterate(xmlKey .. ".fillTypes.fillType", function(_, pFillTypeKey)
            local fillTypeName = THUtils.getXMLValue(xmlFile, pFillTypeKey, "#name")
            local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
            if fillTypeInfo == nil or fillTypeInfo.index == FillType.UNKNOWN then
                THUtils.xmlMsg(pFillTypeKey, "name", fillTypeName, fillTypeName == nil, THXMLMessage.INVALID_FILL_TYPE)
            else
                if not self:loadFillTypeMoisture(fillTypeInfo.index, xmlFile, pFillTypeKey) then
                    loadSuccess = false
                    return false
                end
            end
        end)
    end
    if loadSuccess then
        xmlFile:iterate(xmlKey .. ".fruitTypes.fruitType", function(_, pFruitTypeKey)
            local fruitTypeName = THUtils.getXMLValue(xmlFile, pFruitTypeKey, "#name")
            local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitTypeName)
            if fruitTypeInfo == nil or fruitTypeInfo.index == FruitType.UNKNOWN then
                THUtils.xmlMsg(pFruitTypeKey, "name", fruitTypeName, fruitTypeName == nil, THXMLMessage.INVALID_FRUIT_TYPE)
            else
                if not self:loadFruitTypeMoisture(fruitTypeInfo.index, xmlFile, pFruitTypeKey) then
                    loadSuccess = false
                    return false
                end
            end
        end)
    end
    return loadSuccess
end
function THMoistureSystem:loadFillTypeMoisture(fillType, xmlFile, xmlKey)
    local thFillTypeManager = self.thFillTypeManager
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType, true) -- force
    local minPercent = THUtils.getXMLValue(xmlFile, xmlKey, "#min")
    local maxPercent = THUtils.getXMLValue(xmlFile, xmlKey, "#max")
    local defPercent = THUtils.getXMLValue(xmlFile, xmlKey, "#default")
    local loadSuccess = true
    if fillTypeMoistureData == nil then
        local fillTypeInfo = thFillTypeManager:getFillType(fillType)
        if fillTypeInfo ~= nil then
            THUtils.warningMsg("Could not find moisture data for fillType %q [%s]", fillTypeInfo.name, fillTypeInfo.index)
        end
    elseif fillTypeMoistureData.isEnabled then
        local isMoistureValid = false
        if minPercent ~= nil or maxPercent ~= nil then
            minPercent = THUtils.getNoNil(minPercent, 0)
            maxPercent = THUtils.getNoNil(maxPercent, 100)
            isMoistureValid = true -- activate moisture for this fillType
            if minPercent < 0 or minPercent > 100 then
                THUtils.xmlMsg(xmlKey, "min", minPercent, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                isMoistureValid = false
            elseif maxPercent < 0 or maxPercent > 100 then
                THUtils.xmlMsg(xmlKey, "max", maxPercent, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                isMoistureValid = false
            elseif minPercent > maxPercent then
                THUtils.xmlMsg(xmlKey, "max", maxPercent, true, "is less than min (%s)", minPercent)
                isMoistureValid = false
            else
                fillTypeMoistureData.minPercent = MathUtil.clamp(minPercent / 100, 0, 1)
                fillTypeMoistureData.maxPercent = MathUtil.clamp(maxPercent / 100, 0, 1)
            end
        end
        local xmlMinPercent = fillTypeMoistureData.minPercent * 100
        local xmlMaxPercent = fillTypeMoistureData.maxPercent * 100
        if isMoistureValid then
            if defPercent ~= nil then
                if defPercent < 0 or defPercent > 100 then
                    THUtils.xmlMsg(xmlKey, "default", defPercent, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                    isMoistureValid = false
                elseif defPercent < xmlMinPercent then
                    THUtils.xmlMsg(xmlKey, "default", defPercent, true, "is less than minimum moisture (%s)", xmlMinPercent)
                    isMoistureValid = false
                elseif defPercent > xmlMaxPercent then
                    THUtils.xmlMsg(xmlKey, "default", defPercent, true, "is greater than maximum moisture (%s)", xmlMaxPercent)
                    isMoistureValid = false
                else
                    fillTypeMoistureData.defPercent = MathUtil.clamp(defPercent / 100, 0, 1)
                end
            end
            if not isMoistureValid or defPercent == nil then
                fillTypeMoistureData.defPercent = (fillTypeMoistureData.minPercent + fillTypeMoistureData.maxPercent) / 2
            end
        end
        if isMoistureValid then
            local moistureFactors = fillTypeMoistureData.factors
            local isCurveReset = false
            xmlFile:iterate(xmlKey .. ".factors.factor", function(_, pFactorKey)
                local moisture = THUtils.getXMLValue(xmlFile, pFactorKey, "#moisture")
                local priceFactor = THUtils.getXMLValue(xmlFile, pFactorKey, "#price", 1)
                local yieldFactor = THUtils.getXMLValue(xmlFile, pFactorKey, "#yield", 1)
                local decayFactor = THUtils.getXMLValue(xmlFile, pFactorKey, "#decay", 0)
                if moisture == nil then
                    THUtils.xmlMsg(pFactorKey, "moisture", moisture, true, THXMLMessage.INVALID_VALUE)
                elseif moisture < 0 or moisture > 100 then
                    THUtils.xmlMsg(pFactorKey, "moisture", moisture, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                elseif priceFactor < 0 then
                    THUtils.xmlMsg(pFactorKey, "price", priceFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                elseif yieldFactor < 0 then
                    THUtils.xmlMsg(pFactorKey, "yield", yieldFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                elseif decayFactor < 0 or decayFactor > 1 then
                    THUtils.xmlMsg(pFactorKey, "decay", decayFactor, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 1)
                else
                    local realMoisture = MathUtil.clamp(moisture / 100, 0, 1)
                    if moisture < xmlMinPercent then
                        THUtils.xmlMsg(pFactorKey, "moisture", moisture, true, "is less than minimum moisture (%s)", xmlMinPercent)
                    elseif moisture > xmlMaxPercent then
                        THUtils.xmlMsg(pFactorKey, "moisture", moisture, true, "is greater than maximum moisture (%s)", xmlMaxPercent)
                    else
                        if not isCurveReset then
                            THUtils.clearTable(moistureFactors.curve)
                            isCurveReset = true
                        end
                        if not self:setFillTypeMoistureFactors(fillTypeMoistureData.fillType.index, realMoisture, priceFactor, yieldFactor, decayFactor) then
                            THUtils.stackMsg("Could not set moisture factors for fillType %q [%s]", fillTypeMoistureData.fillType.name, fillTypeMoistureData.fillType.index)
                            isMoistureValid = false
                            return false
                        end
                    end
                end
            end)
        end
        fillTypeMoistureData.hasMoisture = isMoistureValid == true
    end
    return loadSuccess
end
function THMoistureSystem:loadFruitTypeMoisture(fruitType, xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local pfManager = self.thPFManager
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    local loadSuccess = true
    if fruitTypeMoistureData ~= nil then
        local fruitTypeIndex = fruitTypeMoistureData.fruitType.index
        if xmlFile:hasProperty(xmlKey .. ".dryingStates") then
            THUtils.clearTable(fruitTypeMoistureData.dryingStates)
            xmlFile:iterate(xmlKey .. ".dryingStates.state", function(_, pDryingStateKey)
                local minStateId = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#min")
                local maxStateId = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#max")
                local pctStart = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#pctStart")
                local pctEnd = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#pctEnd")
                local hoursToDry = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#hoursToDry")
                minStateId = THUtils.getNoNil(minStateId, maxStateId, true)
                maxStateId = THUtils.getNoNil(maxStateId, minStateId, true)
                local minState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, minStateId, true)
                local maxState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, maxStateId, true)
                local isDryingStateValid = true
                if minState == nil or minState < 0 then
                    THUtils.xmlMsg(pDryingStateKey, "min", minStateId, true, THXMLMessage.INVALID_VALUE)
                    isDryingStateValid = false
                elseif maxState == nil or maxState < 0 then
                    THUtils.xmlMsg(pDryingStateKey, "max", maxStateId, true, THXMLMessage.INVALID_VALUE)
                    isDryingStateValid = false
                end
                if isDryingStateValid then
                    if pctStart == nil then
                        THUtils.xmlMsg(pDryingStateKey, "pctStart", pctStart, true, THXMLMessage.INVALID_VALUE)
                        isDryingStateValid = false
                    elseif pctStart < 0 or pctStart > 100 then
                        THUtils.xmlMsg(pDryingStateKey, "pctStart", pctStart, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                        isDryingStateValid = false
                    else
                        pctStart = pctStart / 100
                    end
                end
                if isDryingStateValid then
                    if pctEnd == nil then
                        THUtils.xmlMsg(pDryingStateKey, "pctEnd", pctEnd, true, THXMLMessage.INVALID_VALUE)
                        isDryingStateValid = false
                    elseif pctEnd < 0 or pctEnd > 100 then
                        THUtils.xmlMsg(pDryingStateKey, "pctEnd", pctEnd, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                        isDryingStateValid = false
                    else
                        pctEnd = pctEnd / 100
                    end
                end
                if isDryingStateValid then
                    local dryingStateData = self:createFruitTypeDryingState(fruitTypeIndex, minState, maxState, pctStart, pctEnd)
                    if dryingStateData == nil then
                        loadSuccess = false
                        return false
                    end
                    if hoursToDry ~= nil then
                        if hoursToDry < 0 then
                            THUtils.xmlMsg(pDryingStateKey, "hoursToDry", hoursToDry, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                            isDryingStateValid = false
                        else
                            dryingStateData.timeToDry = hoursToDry * 60 -- convert to minutes to dry
                        end
                    end
                    local resetPeriods = dryingStateData.resetPeriods
                    local resetPeriodsKey = pDryingStateKey .. ".resetPeriods"
                    if isDryingStateValid and xmlFile:hasProperty(resetPeriodsKey) then
                        resetPeriods = THUtils.clearTable(dryingStateData.resetPeriods)
                        xmlFile:iterate(resetPeriodsKey .. ".period", function(_, pResetPeriodKey)
                            local resetPeriodId = THUtils.getXMLValue(xmlFile, pResetPeriodKey, "#id")
                            local isResetPeriod = THUtils.getXMLValue(xmlFile, pResetPeriodKey, "#reset", true)
                            if resetPeriodId == nil then
                                THUtils.xmlMsg(pResetPeriodKey, "id", resetPeriodId, true, THXMLMessage.INVALID_VALUE)
                            else
                                local period = tonumber(resetPeriodId)
                                if period == nil or self.resetPeriodIndexToId[period] == nil then
                                    period = Environment.PERIOD[resetPeriodId:upper()]
                                end
                                if period == nil then
                                    THUtils.xmlMsg(pResetPeriodKey, "id", resetPeriodId, true, THXMLMessage.INVALID_VALUE)
                                elseif isResetPeriod then
                                    resetPeriods[period] = true
                                else
                                    resetPeriods[period] = nil
                                end
                            end
                        end)
                        if next(resetPeriods) ~= nil then
                            dryingStateData.hasResetPeriods = true
                        end
                    end
                    local dryingFactors = dryingStateData.dryingFactors
                    if isDryingStateValid then
                        local dryingFactorsKey = pDryingStateKey .. ".baseFactors"
                        dryingFactors.night = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#night", THMoistureSystem.DEFAULT_DRYING_FACTORS.NIGHT)
                        dryingFactors.cloud = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#cloud", THMoistureSystem.DEFAULT_DRYING_FACTORS.CLOUD)
                        dryingFactors.rain = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#rain", THMoistureSystem.DEFAULT_DRYING_FACTORS.RAIN)
                        dryingFactors.storm = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#storm", THMoistureSystem.DEFAULT_DRYING_FACTORS.STORM)
                        dryingFactors.snow = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#snow", THMoistureSystem.DEFAULT_DRYING_FACTORS.SNOW)
                        dryingFactors.wind = THUtils.getXMLValue(xmlFile, dryingFactorsKey, "#wind", THMoistureSystem.DEFAULT_DRYING_FACTORS.WIND)
                    end
                    if isDryingStateValid then
                        local humidityFactorsKey = pDryingStateKey .. ".humidityFactors"
                        if xmlFile:hasProperty(humidityFactorsKey) then
                            local humidityFactors = THUtils.clearTable(dryingFactors.humidity)
                            xmlFile:iterate(humidityFactorsKey .. ".humidity", function(_, pHumidityFactorKey)
                                local humidityValue = THUtils.getXMLValue(xmlFile, pHumidityFactorKey, "#value")
                                local humidityFactor = THUtils.getXMLValue(xmlFile, pHumidityFactorKey, "#factor", 1)
                                if humidityValue == nil or humidityValue < 0 or humidityValue > 100 then
                                    THUtils.xmlMsg(pHumidityFactorKey, "value", humidityValue, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                                else
                                    local humidityFactorData = {
                                        value = humidityValue / 100,
                                        factor = humidityFactor
                                    }
                                    local isHumidityValueAdded = false
                                    for _, otherFactorData in pairs(humidityFactors) do
                                        if otherFactorData.value == humidityFactorData.value then
                                            isHumidityValueAdded = true
                                            break
                                        end
                                    end
                                    if not isHumidityValueAdded then
                                        table.insert(humidityFactors, humidityFactorData)
                                    end
                                end
                            end)
                            if #humidityFactors > 0 then
                                table.sort(humidityFactors, function(pa, pb)
                                    return pa.value < pb.value
                                end)
                            end
                        end
                    end
                    if isDryingStateValid then
                        local soilTypeFactorsKey = pDryingStateKey .. ".soilTypeFactors"
                        if xmlFile:hasProperty(soilTypeFactorsKey) then
                            local soilTypeFactors = THUtils.clearTable(dryingFactors.soilTypes)
                            xmlFile:iterate(soilTypeFactorsKey .. ".soilType", function(_, pSoilTypeFactorKey)
                                local soilTypeIndex = THUtils.getXMLValue(xmlFile, pSoilTypeFactorKey, "#index")
                                local soilTypeFactor = THUtils.getXMLValue(xmlFile, pSoilTypeFactorKey, "#factor", 1)
                                local soilTypeInfo = pfManager:getSoilType(soilTypeIndex)
                                if soilTypeInfo == nil then
                                    THUtils.xmlMsg(pSoilTypeFactorKey, "index", soilTypeIndex, false, THXMLMessage.INVALID_VALUE)
                                elseif soilTypeFactor < 0 then
                                    THUtils.xmlMsg(pSoilTypeFactorKey, "factor", soilTypeFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                                else
                                    soilTypeFactors[soilTypeIndex] = soilTypeFactor
                                end
                            end)
                        end
                    end
                    if isDryingStateValid then
                        dryingStateData.index = #fruitTypeMoistureData.dryingStates + 1
                        fruitTypeMoistureData.dryingStates[dryingStateData.index] = dryingStateData
                    end
                end
            end)
        end
    end
    return loadSuccess
end
function THMoistureSystem:loadFromSavegameXMLFile(xmlFile, xmlKey)
    local loadSuccess = THMoistureSystem:superClass().loadFromSavegameXMLFile(self, xmlFile, xmlKey)
    if self.isServer and loadSuccess then
        local thFruitTypeManager = self.thFruitTypeManager
        self.dryingStateCachedTime = THUtils.getXMLValue(xmlFile, xmlKey, "#cachedDryingTime", 0)
        self.periodLength = THUtils.getXMLValue(xmlFile, xmlKey, "#periodLength", 1)
        local dryingConditions = self.dryingConditions
        local dryingConditionsKey = xmlKey .. ".dryingConditions"
        if xmlFile:hasProperty(dryingConditionsKey) then
            dryingConditions.isRaining = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#isRaining", false)
            dryingConditions.isDayTime = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#isDayTime", false)
            dryingConditions.temperature = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#temp", 0)
            dryingConditions.dewPoint = THUtils.getXMLValue(xmlFile, dryingConditionsKey, "#dewPoint", 0)
            local humidity = THMoistureUtil.calculateHumidity(dryingConditions.temperature, dryingConditions.dewPoint)
            dryingConditions.humidity = MathUtil.clamp(humidity, dryingConditions.absMinHumidity, dryingConditions.absMaxHumidity)
            local farmlandFactorsKey = string.format("%s.farmlandFactors", dryingConditionsKey)
            THUtils.clearTable(dryingConditions.farmlandFactors)
            xmlFile:iterate(farmlandFactorsKey .. ".farmland", function(_, pFarmlandFactorKey)
                local farmlandId = THUtils.getXMLValue(xmlFile, pFarmlandFactorKey, "#id")
                local farmlandFactor = THUtils.getXMLValue(xmlFile, pFarmlandFactorKey, "#factor")
                if farmlandId ~= nil then
                    self:setFarmlandDryingFactor(farmlandId, farmlandFactor, true)
                end
            end)
            local currentRainData = dryingConditions.currentRainData
            local currentRainKey = dryingConditionsKey .. ".currentRain"
            if xmlFile:hasProperty(currentRainKey) then
                currentRainData.startDay = THUtils.getXMLValue(xmlFile, currentRainKey, "#startDay", 1)
                currentRainData.startTime = THUtils.getXMLValue(xmlFile, currentRainKey, "#startTime", 0)
                currentRainData.endDay = THUtils.getXMLValue(xmlFile, currentRainKey, "#endDay", 1)
                currentRainData.endTime = THUtils.getXMLValue(xmlFile, currentRainKey, "#endTime", 0)
                currentRainData.duration = THUtils.getXMLValue(xmlFile, currentRainKey, "#duration", 0)
            end
            local lastRainData = dryingConditions.lastRainData
            local lastRainKey = dryingConditionsKey .. ".lastRain"
            if xmlFile:hasProperty(lastRainKey) then
                lastRainData.startDay = THUtils.getXMLValue(xmlFile, lastRainKey, "#startDay", 1)
                lastRainData.startTime = THUtils.getXMLValue(xmlFile, lastRainKey, "#startTime", 0)
                lastRainData.endDay = THUtils.getXMLValue(xmlFile, lastRainKey, "#endDay", 1)
                lastRainData.endTime = THUtils.getXMLValue(xmlFile, lastRainKey, "#endTime", 0)
                lastRainData.duration = THUtils.getXMLValue(xmlFile, lastRainKey, "#duration", 0)
            end
            dryingConditions.isInitialized = true
        end
        xmlFile:iterate(xmlKey .. ".fruitTypes.fruitType", function(_, pFruitTypeKey)
            local fruitTypeName = THUtils.getXMLValue(xmlFile, pFruitTypeKey, "#name")
            local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitTypeName, true) -- force
            if fruitTypeMoistureData == nil then
                local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitTypeName)
                if fruitTypeInfo == nil or fruitTypeInfo.index == FruitType.UNKNOWN then
                    THUtils.xmlMsg(pFruitTypeKey, "name", fruitTypeName, false, THXMLMessage.INVALID_FRUIT_TYPE)
                else
                    THUtils.xmlMsg(pFruitTypeKey, "name", fruitTypeName, true, "Could not load moisture values")
                end
            elseif fruitTypeMoistureData.isEnabled then
                xmlFile:iterate(pFruitTypeKey .. ".dryingStates.dryingState", function(_, pDryingStateKey)
                    local dryingStateIndex = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#index")
                    if dryingStateIndex ~= nil and fruitTypeMoistureData.dryingStates[dryingStateIndex] ~= nil then
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        local dryingTimes = dryingStateData.dryingTimes
                        local dryingTimesKey = pDryingStateKey .. ".dryingTimes"
                        dryingStateData.averageTemp = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#averageTemp", 0)
                        dryingStateData.averageWind = THUtils.getXMLValue(xmlFile, pDryingStateKey, "#averageWind", 0)
                        dryingTimes.total = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#total", 0)
                        dryingTimes.night = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#night", 0)
                        dryingTimes.cloud = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#cloud", 0)
                        dryingTimes.rain = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#rain", 0)
                        dryingTimes.storm = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#storm", 0)
                        dryingTimes.snow = THUtils.getXMLValue(xmlFile, dryingTimesKey, "#snow", 0)
                    end
                end)
            end
        end)
    end
    return loadSuccess
end
function THMoistureSystem:saveToSavegameXMLFile(xmlFile, xmlKey)
    local thFruitTypeManager = self.thFruitTypeManager
    local saveSuccess = THMoistureSystem:superClass().saveToSavegameXMLFile(self, xmlFile, xmlKey)
    if self.isServer and saveSuccess then
        xmlFile:setValue(xmlKey .. "#cachedDryingTime", self.dryingStateCachedTime)
        xmlFile:setValue(xmlKey .. "#periodLength", self.periodLength)
        local dryingConditions = self.dryingConditions
        local dryingConditionsKey = xmlKey .. ".dryingConditions"
        xmlFile:setValue(dryingConditionsKey .. "#isRaining", dryingConditions.isRaining)
        xmlFile:setValue(dryingConditionsKey .. "#isDayTime", dryingConditions.isDayTime)
        xmlFile:setValue(dryingConditionsKey .. "#temp", dryingConditions.temperature)
        xmlFile:setValue(dryingConditionsKey .. "#dewPoint", dryingConditions.dewPoint)
        if next(dryingConditions.farmlandFactors) ~= nil then
            local farmlandXMLIndex = 0
            for farmlandId, farmlandFactor in pairs(dryingConditions.farmlandFactors) do
                local farmlandFactorKey = string.format("%s.farmlandFactors.farmland(%d)", dryingConditionsKey, farmlandXMLIndex)
                xmlFile:setValue(farmlandFactorKey .. "#id", farmlandId)
                xmlFile:setValue(farmlandFactorKey .. "#factor", farmlandFactor)
                farmlandXMLIndex = farmlandXMLIndex + 1
            end
        end
        local currentRainData = dryingConditions.currentRainData
        local lastRainData = dryingConditions.lastRainData
        if currentRainData.duration > 0 then
            local currentRainKey = dryingConditionsKey .. ".currentRain"
            xmlFile:setValue(currentRainKey .. "#startDay", currentRainData.startDay)
            xmlFile:setValue(currentRainKey .. "#startTime", currentRainData.startTime)
            xmlFile:setValue(currentRainKey .. "#endDay", currentRainData.endDay)
            xmlFile:setValue(currentRainKey .. "#endTime", currentRainData.endTime)
            xmlFile:setValue(currentRainKey .. "#duration", currentRainData.duration)
        end
        if currentRainData.duration > 0 then
            local lastRainKey = dryingConditionsKey .. ".lastRain"
            xmlFile:setValue(lastRainKey .. "#startDay", lastRainData.startDay)
            xmlFile:setValue(lastRainKey .. "#startTime", lastRainData.startTime)
            xmlFile:setValue(lastRainKey .. "#endDay", lastRainData.endDay)
            xmlFile:setValue(lastRainKey .. "#endTime", lastRainData.endTime)
            xmlFile:setValue(lastRainKey .. "#duration", lastRainData.duration)
        end
        local numFruitTypes = #self.fruitTypes
        if numFruitTypes > 0 then
            local fruitTypeXMLIndex = 0
            for fruitTypeIndex, fruitTypeMoistureData in ipairs(self.fruitTypes) do
                local fruitTypeKey = string.format("%s.fruitTypes.fruitType(%d)", xmlKey, fruitTypeXMLIndex)
                local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitTypeIndex)
                if fruitTypeInfo ~= nil and fruitTypeInfo.index ~= FruitType.UNKNOWN then
                    xmlFile:setValue(fruitTypeKey .. "#name", fruitTypeInfo.name)
                    local numDryingStates = #fruitTypeMoistureData.dryingStates
                    if numDryingStates > 0 then
                        local dryingStateXMLIndex = 0
                        local dryingStateKey = string.format("%s.dryingStates.dryingState(%d)", fruitTypeKey, dryingStateXMLIndex)
                        for dryingStateIndex, dryingStateData in ipairs(fruitTypeMoistureData.dryingStates) do
                            xmlFile:setValue(dryingStateKey .. "#index", dryingStateIndex)
                            xmlFile:setValue(dryingStateKey .. "#averageTemp", dryingStateData.averageTemp)
                            xmlFile:setValue(dryingStateKey .. "#averageWind", dryingStateData.averageWind)
                            local dryingTimesKey = dryingStateKey .. ".dryingTimes"
                            local dryingTimes = dryingStateData.dryingTimes
                            xmlFile:setValue(dryingTimesKey .. "#total", dryingTimes.total)
                            xmlFile:setValue(dryingTimesKey .. "#night", dryingTimes.night)
                            xmlFile:setValue(dryingTimesKey .. "#cloud", dryingTimes.cloud)
                            xmlFile:setValue(dryingTimesKey .. "#rain", dryingTimes.rain)
                            xmlFile:setValue(dryingTimesKey .. "#storm", dryingTimes.storm)
                            xmlFile:setValue(dryingTimesKey .. "#snow", dryingTimes.snow)
                            dryingStateXMLIndex = dryingStateXMLIndex + 1
                        end
                    end
                    fruitTypeXMLIndex = fruitTypeXMLIndex + 1
                end
            end
        end
    end
    return saveSuccess
end
function THMoistureSystem:onPreInit()
    if not self.isPreInitFinished then
        THMoistureSystem:superClass().onPreInit(self)
        self:initSettings()
        self.isPreInitFinished = true
    end
end
function THMoistureSystem:onPostInit()
    if not self.isPostInitFinished then
        THMoistureSystem:superClass().onPostInit(self)
        self.isPostInitFinished = true
    end
end
function THMoistureSystem:onLoadEnvironment(environment, mission)
    THMoistureSystem:superClass().onLoadEnvironment(self, environment, mission)
    self.environment = environment
    self.growthSystem = mission.growthSystem
    self.weatherSystem = self.environment.weather
    self.windUpdater = self.weatherSystem.windUpdater
end
function THMoistureSystem:onLoadMapFinished(mission)
    THMoistureSystem:superClass().onLoadMapFinished(self, mission)
    self:initializeDataTables()
    self:loadConfiguration()
    self:finalizeDataTables()
    self:loadFromSavegame()
end
function THMoistureSystem:onUpdateTick(dt)
    THMoistureSystem:superClass().onUpdateTick(self, dt)
    if self.isServer then
        self.isDryingStateUpdateSafe = true
        if not self.dryingConditions.isInitialized then
            self:updateDryingConditions()
        end
        if not self.areDryingStatesInitialized then
            self:updateFruitTypeDryingStates(0)
            self.areDryingStatesInitialized = true
        end
    end
end
function THMoistureSystem:onReadStream(streamId, connection)
    if connection:getIsServer() then
        self:readMoistureFromStream(streamId, connection, true)
    end
end
function THMoistureSystem:onWriteStream(streamId, connection)
    if not connection:getIsServer() then
        self:writeMoistureToStream(streamId, connection, true)
    end
end
function THMoistureSystem:onReadUpdateStream(streamId, timeStamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then -- isDirty
            self:readMoistureFromStream(streamId, connection)
        end
    end
end
function THMoistureSystem:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.dirtyFlag) ~= 0) then
            self:writeMoistureToStream(streamId, connection)
        end
    end
end
function THMoistureSystem:readMoistureFromStream(streamId, connection, isFullSync)
    if connection:getIsServer() then
        local farmlandManager = self.farmlandManager
        local dryingConditions = self.dryingConditions
        local maxFactorBitValue = (2 ^ THMoistureSystem.FACTOR_SEND_BITS) - 1
        self.periodLength = streamReadUIntN(streamId, THMoistureSystem.PERIOD_SEND_BITS)
        if streamReadBool(streamId) then -- dryingConditions initialized
            local sentHumidity = streamReadUIntN(streamId, THMoistureSystem.FACTOR_SEND_BITS)
            dryingConditions.humidity = MathUtil.clamp(sentHumidity / maxFactorBitValue, 0, 1)
            local farmlandsArray = farmlandManager.sortedFarmlandIds
            if farmlandsArray ~= nil then
                local numFarmlands = #farmlandsArray
                if numFarmlands > 0 then
                    for farmlandId = 1, numFarmlands do
                        local sentFactor = streamReadUIntN(streamId, THMoistureSystem.FACTOR_SEND_BITS)
                        dryingConditions.farmlandFactors[farmlandId] = MathUtil.clamp(sentFactor / maxFactorBitValue, 0, 1)
                    end
                end
            end
            dryingConditions.isInitialized = true
        end
        local numFruitTypes = #self.fruitTypes
        if numFruitTypes > 0 then
            for i = 1, numFruitTypes do
                local fruitTypeMoistureData = self.fruitTypes[i]
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        local dryingTimes = dryingStateData.dryingTimes
                        dryingTimes.total = streamReadFloat32(streamId)
                    end
                end
            end
        end
    end
end
function THMoistureSystem:writeMoistureToStream(streamId, connection, isFullSync)
    if not connection:getIsServer() then
        local farmlandManager = self.farmlandManager
        local dryingConditions = self.dryingConditions
        local maxFactorBitValue = (2 ^ THMoistureSystem.FACTOR_SEND_BITS) - 1
        streamWriteUIntN(streamId, self.periodLength, THMoistureSystem.PERIOD_SEND_BITS)
        if streamWriteBool(streamId, dryingConditions.isInitialized == true) then
            local sentHumidity = math.floor(dryingConditions.humidity * maxFactorBitValue)
            streamWriteUIntN(streamId, sentHumidity, THMoistureSystem.FACTOR_SEND_BITS)
            local farmlandsArray = farmlandManager.sortedFarmlandIds
            if farmlandsArray ~= nil then
                local numFarmlands = #farmlandsArray
                for farmlandId = 1, numFarmlands do
                    local farmlandFactor = dryingConditions.farmlandFactors[farmlandId] or 1
                    local sentFactor = math.floor(farmlandFactor * maxFactorBitValue)
                    streamWriteUIntN(streamId, sentFactor, THMoistureSystem.FACTOR_SEND_BITS)
                end
            end
        end
        local numFruitTypes = #self.fruitTypes
        if numFruitTypes > 0 then
            for i = 1, numFruitTypes do
                local fruitTypeMoistureData = self.fruitTypes[i]
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        local dryingTimes = dryingStateData.dryingTimes
                        streamWriteFloat32(streamId, dryingTimes.total)
                    end
                end
            end
        end
    end
end
function THMoistureSystem:onMinuteChanged(currentMinute)
    if self.isServer then
        if not self.dryingConditions.isInitialized then
            self:updateDryingConditions()
        end
        self:updateFruitTypeDryingStates(1)
    end
end
function THMoistureSystem:onHourChanged(currentHour)
    if self.isServer then
        self:updateDryingConditions()
    end
end
function THMoistureSystem:onDayChanged(currentDay)
    if self.isServer then
        local numMoistureFruitTypes = #self.fruitTypes
        if numMoistureFruitTypes > 0 then
            for fruitTypeIndex = 1, numMoistureFruitTypes do
                local fruitTypeMoistureData = self.fruitTypes[fruitTypeIndex]
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        if not dryingStateData.hasResetPeriods then
                            self:resetFruitTypeDryingState(fruitTypeIndex, dryingStateIndex)
                        end
                    end
                end
            end
        end
    end
end
function THMoistureSystem:onPeriodChanged(currentPeriod)
    if self.isServer then
        local environment = self.environment
        local numMoistureFruitTypes = #self.fruitTypes
        self.periodLength = THUtils.getNoNil(environment.daysPerPeriod, 1)
        if numMoistureFruitTypes > 0 then
            for fruitTypeIndex = 1, numMoistureFruitTypes do
                local fruitTypeMoistureData = self.fruitTypes[fruitTypeIndex]
                local numDryingStates = #fruitTypeMoistureData.dryingStates
                if numDryingStates > 0 then
                    for dryingStateIndex = 1, numDryingStates do
                        local dryingStateData = fruitTypeMoistureData.dryingStates[dryingStateIndex]
                        if dryingStateData.hasResetPeriods
                            and dryingStateData.resetPeriods[currentPeriod]
                        then
                            self:resetFruitTypeDryingState(fruitTypeIndex, dryingStateIndex)
                        end
                    end
                end
            end
        end
        self:raiseDirtyFlags(self.dirtyFlag)
    end
end
function THMoistureSystem:onWeatherChanged()
    if self.isServer then
        self:updateDryingConditions()
    end
end
function THMoistureSystem:getCurrentDayTime()
    local weatherSystem = self.weatherSystem
    local currentTime = 0
    local currentDay = 1
    if weatherSystem ~= nil and weatherSystem.owner ~= nil then
        currentTime = THUtils.getNoNil(weatherSystem.owner.dayTime, 0)
        currentDay = THUtils.getNoNil(weatherSystem.owner.currentMonotonicDay, 1)
    end
    return currentTime, currentDay
end
function THMoistureSystem:getCurrentWeatherData()
    local weatherSystem = self.weatherSystem
    if weatherSystem ~= nil and weatherSystem.forecastItems ~= nil then
        local currentTime, currentDay = self:getCurrentDayTime()
        local numForecastItems = #weatherSystem.forecastItems
        if numForecastItems > 0 then
            for itemIndex = 1, numForecastItems do
                local forecastItem = weatherSystem.forecastItems[itemIndex]
                if forecastItem.startDay > currentDay
                    or (forecastItem.startDay == currentDay and forecastItem.startDayTime > currentTime)
                then
                    if itemIndex > 1 then
                        local lastForecastItem = weatherSystem.forecastItems[itemIndex - 1]
                        local weatherObject = weatherSystem:getWeatherObjectByIndex(lastForecastItem.season, lastForecastItem.objectIndex)
                        if weatherObject ~= nil then
                            return weatherObject.weatherType.index, weatherObject, lastForecastItem
                        end
                    end
                end
            end
        end
    end
    return WeatherType.SUN
end
function THMoistureSystem:getNextWeatherData()
    local weatherSystem = self.weatherSystem
    if weatherSystem ~= nil and weatherSystem.forecastItems ~= nil then
        local currentTime, currentDay = self:getCurrentDayTime()
        local numForecastItems = #weatherSystem.forecastItems
        if numForecastItems > 0 then
            for itemIndex = 1, numForecastItems do
                local forecastItem = weatherSystem.forecastItems[itemIndex]
                if forecastItem.startDay > currentDay
                    or (forecastItem.startDay == currentDay and forecastItem.startDayTime > currentTime)
                then
                    local weatherObject = weatherSystem:getWeatherObjectByIndex(forecastItem.season, forecastItem.objectIndex)
                    if weatherObject ~= nil then
                        return weatherObject.weatherType.index, weatherObject, forecastItem
                    end
                end
            end
        end
    end
    return WeatherType.SUN
end
function THMoistureSystem:getTimeSinceRain()
    local weatherSystem = self.weatherSystem
    local timeSinceRain = 0
    if weatherSystem ~= nil and weatherSystem.owner ~= nil then
        if not weatherSystem:getIsRaining() then
            timeSinceRain = weatherSystem:getTimeSinceLastRain()
            if timeSinceRain == nil or timeSinceRain < 0 or timeSinceRain == math.huge then
                local currentTime, currentDay = self:getCurrentDayTime()
                local lastRainData = self.dryingConditions.lastRainData
                if lastRainData.endDay > currentDay
                    or (lastRainData.endDay == currentDay and lastRainData.endTime > currentTime)
                then
                    timeSinceRain = 0
                else
                    local absCurrentTime = currentTime + ((currentDay - lastRainData.endDay) * weatherSystem.owner.dayLength)
                    timeSinceRain = (absCurrentTime - lastRainData.endTime) / 60000 -- ms to minutes
                end
            end
        end
    end
    return timeSinceRain
end
function THMoistureSystem:getCurrentHumidityValues()
    local weatherSystem = self.weatherSystem
    local dryingConditions = self.dryingConditions
    if dryingConditions.isInitialized then
        return dryingConditions.humidity, dryingConditions.dewPoint
    else
        local currentTemp = weatherSystem:getCurrentTemperature()
        local absMinHumidity = dryingConditions.absMinHumidity
        local absMaxHumidity = dryingConditions.absMaxHumidity
        local humidity = (absMinHumidity + absMaxHumidity) / 2
        local dewPoint = THMoistureUtil.calculateDewPoint(currentTemp, humidity)
        return humidity, dewPoint
    end
end
function THMoistureSystem:getTimeUntilRain(isAbsolute)
    local weatherSystem = self.weatherSystem
    local timeUntilRain = math.huge
    if weatherSystem ~= nil and weatherSystem.owner ~= nil and weatherSystem.forecastItems ~= nil then
        if weatherSystem:getIsRaining() then
            timeUntilRain = 0
        else
            local currentTime, currentDay = self:getCurrentDayTime()
            local dayLength = weatherSystem.owner.dayLength
            local numForecastItems = #weatherSystem.forecastItems
            if numForecastItems > 0 then
                for itemIndex = 1, numForecastItems do -- start at next forecast item
                    local forecastItem = weatherSystem.forecastItems[itemIndex]
                    local weatherObject = weatherSystem:getWeatherObjectByIndex(forecastItem.season, forecastItem.objectIndex)
                    if weatherObject ~= nil then
                        if forecastItem.startDay > currentDay
                            or (forecastItem.startDay == currentDay and forecastItem.startDayTime > currentTime)
                        then
                            if weatherObject.weatherType.index == WeatherType.RAIN
                                or weatherObject.weatherType.index == WeatherType.SNOW
                            then
                                local absStartTime = forecastItem.startDayTime + ((forecastItem.startDay - currentDay) * dayLength)
                                timeUntilRain = (absStartTime - currentTime) / 60000 -- ms to minutes
                                break
                            elseif not isAbsolute then -- stop after first valid attempt
                                break
                            end
                        end
                    end
                end
            end
        end
    end
    return timeUntilRain
end
function THMoistureSystem:getFarmlandDryingFactor(farmlandId)
    local farmlandManager = self.farmlandManager
    local farmlandFactor = 1
    if farmlandManager ~= nil then
        local farmlandInfo = farmlandManager:getFarmlandById(farmlandId)
        if farmlandInfo ~= nil then
            farmlandFactor = self.dryingConditions.farmlandFactors[farmlandInfo.id] or 1
        end
    end
    return farmlandFactor
end
function THMoistureSystem:setFarmlandDryingFactor(farmlandId, factor, noEventSend)
    if THUtils.argIsValid(type(factor) == "number", "factor", factor) then
        if self.isServer or noEventSend == true then
            local farmlandManager = self.farmlandManager
            if farmlandManager ~= nil and farmlandId ~= nil then
                local farmlandInfo = farmlandManager:getFarmlandById(farmlandId)
                if farmlandInfo ~= nil then
                    factor = MathUtil.clamp(factor, 0, 1)
                    self.dryingConditions.farmlandFactors[farmlandInfo.id] = factor
                    if self.isServer and noEventSend ~= true then
                        self:raiseDirtyFlags(self.dirtyFlag)
                    end
                end
            end
        end
    end
end
function THMoistureSystem:updateDryingConditions(noEventSend)
    if self.isServer then
        local environment = self.environment
        local weatherSystem = self.weatherSystem
        local dryingConditions = self.dryingConditions
        if environment ~= nil and weatherSystem ~= nil and weatherSystem.owner ~= nil then
            local currentTime, currentDay = self:getCurrentDayTime()
            local dayLength = weatherSystem.owner.dayLength
            local curTemp = weatherSystem:getCurrentTemperature()
            local isDayTime = environment.isSunOn
            local isRaining = weatherSystem:getIsRaining()
            local weatherType, _, forecastItem = self:getCurrentWeatherData()
            local nextWeatherType, _, nextForecastItem = self:getNextWeatherData()
            local absMinHumidity = dryingConditions.absMinHumidity
            local absMaxHumidity = dryingConditions.absMaxHumidity
            local absMinDewPoint = THMoistureUtil.calculateDewPoint(curTemp, absMinHumidity)
            local absMaxDewPoint = THMoistureUtil.calculateDewPoint(curTemp, absMaxHumidity)
            local updateFarmlandFactors = false
            absMinDewPoint = math.max(absMinDewPoint, -33)
            absMaxDewPoint = math.min(absMaxDewPoint, 35)
            local timeSinceRain = self:getTimeSinceRain()
            local timeUntilRain = self:getTimeUntilRain(true)
            local currentRainData = dryingConditions.currentRainData
            local lastRainData = dryingConditions.lastRainData
            if not dryingConditions.isInitialized then
                dryingConditions.temperature = curTemp
                dryingConditions.isDayTime = isDayTime
                dryingConditions.isRaining = isRaining
                if isRaining then
                    dryingConditions.humidity = 1
                else
                    dryingConditions.humidity = 0.5
                end
                dryingConditions.dewPoint = THMoistureUtil.calculateDewPoint(dryingConditions.temperature, dryingConditions.humidity)
                lastRainData.startDay = currentDay
                lastRainData.startTime = currentTime
                lastRainData.endDay = currentDay
                lastRainData.endTime = currentTime
                lastRainData.duration = 0
                for key, value in pairs(lastRainData) do
                    currentRainData[key] = value
                end
                updateFarmlandFactors = true
                dryingConditions.isInitialized = true
            end
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THUtils.displayMsg("Current drying conditions:")
                THUtils.displayMsg("- Current temperature: %s", curTemp)
                THUtils.displayMsg("- Abs max dew point: %s", absMaxDewPoint)
                THUtils.displayMsg("- Abs min dew point: %s", absMinDewPoint)
                THUtils.displayMsg("")
            end
            if isRaining or timeSinceRain <= 0 or timeUntilRain <= 0
                or weatherType == WeatherType.RAIN
                or weatherType == WeatherType.SNOW
            then
                if isRaining then -- raining flag specifically set
                    currentRainData.endTime = currentTime
                    currentRainData.endDay = currentDay
                    if isRaining ~= dryingConditions.isRaining then -- new rain event
                        currentRainData.startDay = currentDay
                        currentRainData.startTime = currentTime
                        currentRainData.duration = 0 -- duration calculated at end of event
                    end
                end
                dryingConditions.dewPoint = absMaxDewPoint
            else
                if isRaining ~= dryingConditions.isRaining then -- rain event ended
                    local absEndTime = currentTime
                    if currentDay > currentRainData.startDay then
                        absEndTime = currentRainData.endTime + ((currentDay - currentRainData.startDay) * dayLength)
                    end
                    currentRainData.duration = absEndTime - currentRainData.startTime
                    for key, value in pairs(currentRainData) do
                        lastRainData[key] = value
                    end
                    updateFarmlandFactors = true
                end
                local lastTemp = dryingConditions.temperature
                local tempTrend = curTemp - lastTemp
                local rainFactor = 0
                local weatherFactor = 0
                local dryingTimeFactor = dryingConditions.dryingTimeFactor or 1
                if timeSinceRain < 0 or timeUntilRain < 0 or dryingTimeFactor <= 0 then
                    rainFactor = 1
                else
                    local maxDryingTime = (dryingConditions.maxDryingTime * self.periodLength) / dryingTimeFactor
                    if timeSinceRain ~= math.huge and timeSinceRain < maxDryingTime then
                        rainFactor = 1 - MathUtil.clamp(timeSinceRain / maxDryingTime, 0, 1)
                    end
                    if timeUntilRain ~= math.huge and timeUntilRain < maxDryingTime then
                        local untilRainFactor = 1 - MathUtil.clamp(timeUntilRain / maxDryingTime, 0, 1)
                        rainFactor = math.max(rainFactor, untilRainFactor)
                    end
                end
                if forecastItem ~= nil and nextForecastItem ~= nil then
                    local function getWeatherTypeFactor(pWeatherType)
                        local factor = 0 -- clear sunny weather
                        if pWeatherType ~= WeatherType.SUN then
                            if pWeatherType == WeatherType.RAIN
                                or pWeatherType == WeatherType.SNOW
                            then
                                factor = 1
                            else
                                factor = 0.5
                            end
                        end
                        return factor
                    end
                    local weatherTypeFactor = getWeatherTypeFactor(weatherType)
                    local nextWeatherTypeFactor = getWeatherTypeFactor(nextWeatherType)
                    local weatherStartDay = forecastItem.startDay
                    local weatherStartTime = forecastItem.startDayTime
                    local weatherEndDay = nextForecastItem.startDay
                    local weatherEndTime = nextForecastItem.startDayTime
                    local absCurrentTime = currentTime
                    local absWeatherEndTime = weatherEndTime
                    if currentDay > weatherStartDay then
                        absCurrentTime = currentTime + ((currentDay - weatherStartDay) * dayLength)
                    end
                    if weatherEndDay > weatherStartDay then
                        absWeatherEndTime = weatherEndTime + ((weatherEndDay - weatherStartDay) * dayLength)
                    end
                    if THDebugUtil.getIsEnabled(debugFlagId) then
                        THUtils.displayMsg("- Weather start time: %s", weatherStartTime)
                        THUtils.displayMsg("- Weather end time: %s", absWeatherEndTime)
                        THUtils.displayMsg("- Current weather time: %s", absCurrentTime)
                    end
                    if absCurrentTime >= weatherStartTime and absWeatherEndTime >= absCurrentTime then
                        local weatherTimeAlpha = THUtils.getFactorInRange(weatherStartTime, absWeatherEndTime, absCurrentTime)
                        THDebugUtil.printMsg(debugFlagId, nil, "- Weather time alpha: %s", weatherTimeAlpha)
                        weatherFactor = THUtils.getValueInRange(weatherTypeFactor, nextWeatherTypeFactor, weatherTimeAlpha)
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                end
                local moistureFactor = (weatherFactor + rainFactor) / 2  -- clamped (0 - 1) above
                local lastDewPoint = dryingConditions.dewPoint
                local nextDewPoint = lastDewPoint + (tempTrend * 0.75 * math.random())
                local minDewPoint = math.min(absMaxDewPoint, absMinDewPoint + ((absMaxDewPoint - absMinDewPoint) * moistureFactor))
                local maxDewPoint = absMaxDewPoint
                if tempTrend > 0 then
                    nextDewPoint = lastDewPoint + (tempTrend * 0.75 * moistureFactor * math.random())
                elseif tempTrend < 0 then
                    nextDewPoint = lastDewPoint + (tempTrend * 0.75 * (1 - moistureFactor) * math.random())
                end
                if isRaining ~= dryingConditions.isRaining -- implies rain event ended (isRaining handled above)
                    or (nextDewPoint >= maxDewPoint and lastDewPoint >= maxDewPoint)
                then
                    nextDewPoint = maxDewPoint - ((maxDewPoint - minDewPoint) * (1 - moistureFactor) * math.random())
                elseif nextDewPoint <= minDewPoint and lastDewPoint <= minDewPoint then
                    nextDewPoint = minDewPoint + ((maxDewPoint - minDewPoint) * moistureFactor * math.random())
                end
                dryingConditions.dewPoint = MathUtil.clamp(nextDewPoint, minDewPoint, maxDewPoint)
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    THUtils.displayMsg("- Current weather type: %s", weatherType)
                    THUtils.displayMsg("- Next weather type: %s", nextWeatherType)
                    THUtils.displayMsg("- Weather factor : %s", weatherFactor)
                    THUtils.displayMsg("")
                    THUtils.displayMsg("- Last temperature: %s", lastTemp)
                    THUtils.displayMsg("- Last dew point: %s", lastDewPoint)
                    THUtils.displayMsg("- Adjusted min dew point: %s", minDewPoint)
                    THUtils.displayMsg("")
                    THUtils.displayMsg("- Time since rain: %s", timeSinceRain)
                    THUtils.displayMsg("- Time until rain: %s", timeUntilRain)
                    THUtils.displayMsg("- Rain factor: %s", rainFactor)
                    THUtils.displayMsg("")
                end
            end
            local nextHumidity = THMoistureUtil.calculateHumidity(curTemp, dryingConditions.dewPoint)
            dryingConditions.temperature = curTemp
            dryingConditions.isDayTime = isDayTime
            dryingConditions.isRaining = isRaining
            dryingConditions.humidity = MathUtil.clamp(nextHumidity, absMinHumidity, absMaxHumidity)
            if updateFarmlandFactors then
                local farmlandList = self.farmlandManager:getFarmlands()
                THUtils.clearTable(dryingConditions.farmlandFactors)
                if farmlandList ~= nil then
                    for _, farmlandInfo in pairs(farmlandList) do
                        local farmlandFactor = math.random()
                        self:setFarmlandDryingFactor(farmlandInfo.id, farmlandFactor, true)
                    end
                end
            end
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THUtils.displayMsg("Current Humidity: %0.1f%%", dryingConditions.humidity * 100)
                THUtils.printTable(dryingConditions, 1)
            end
            if noEventSend ~= true then
                self:raiseDirtyFlags(self.dirtyFlag)
            end
        end
    end
end
function THMoistureSystem:getFruitTypeMoistureData(fruitType, force)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitType)
    if fruitTypeInfo ~= nil and fruitTypeInfo.index ~= FruitType.UNKNOWN then
        local fruitTypeMoistureData = self.fruitTypes[fruitTypeInfo.index]
        if fruitTypeMoistureData ~= nil then
            if fruitTypeMoistureData.isEnabled or force == true then
                return fruitTypeMoistureData
            end
        end
    end
end
function THMoistureSystem:getFruitTypeDryingState(fruitType, dryingState)
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    if fruitTypeMoistureData ~= nil and dryingState ~= nil then
        return fruitTypeMoistureData.dryingStates[dryingState]
    end
end
function THMoistureSystem:getFruitTypeDryingStateByGrowthState(fruitType, growthStateId)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    local growthState = thFruitTypeManager:getFruitTypeGrowthState(fruitType, growthStateId, true)
    if growthState == nil or growthState < 0 then
        THDebugUtil.printMsg(debugFlagId, nil, THMessage.ARGUMENT_INVALID, "growthStateId", growthStateId)
    elseif fruitTypeMoistureData ~= nil then
        local dryingState = fruitTypeMoistureData.dryingStateMapping[growthState]
        if dryingState ~= nil then
            return fruitTypeMoistureData.dryingStates[dryingState], growthState
        end
    end
    return nil, -1
end
function THMoistureSystem:createFruitTypeDryingState(fruitType, minStateId, maxStateId, pctStart, pctEnd)
    local thFruitTypeManager = self.thFruitTypeManager
    local fruitTypeMoistureData = self:getFruitTypeMoistureData(fruitType)
    if fruitTypeMoistureData ~= nil then
        local fruitTypeData = fruitTypeMoistureData.fruitType
        local fruitTypeIndex = fruitTypeData.index
        local dryingStateData = {}
        dryingStateData.parent = fruitTypeMoistureData
        local minGrowthState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, minStateId, true)
        local maxGrowthState = thFruitTypeManager:getFruitTypeGrowthState(fruitTypeIndex, maxStateId, true)
        if minGrowthState == nil or minGrowthState < 0 then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "minStateId", minStateId)
        elseif maxGrowthState == nil or maxGrowthState < 0 then
            THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "maxStateId", maxStateId)
        elseif pctStart == nil or pctStart < 0 or pctStart > 1 then
            THUtils.stackMsg("Argument pctStart (%s) must be between 0 and 1", pctStart)
        elseif pctEnd == nil or pctEnd < 0 or pctEnd > 1 then
            THUtils.stackMsg("Argument pctEnd (%s) must be between 0 and 1", pctEnd)
        else
            dryingStateData.minGrowthState = math.min(minGrowthState, maxGrowthState)
            dryingStateData.maxGrowthState = math.max(minGrowthState, maxGrowthState)
            dryingStateData.timeToDry = 360 -- 6 hours to dry
            dryingStateData.averageWind = 0
            dryingStateData.averageTemp = 0
            dryingStateData.startPercent = pctStart
            dryingStateData.endPercent = pctEnd
            dryingStateData.dryingTimes = {
                total = 0,
                night = 0,
                cloud = 0,
                rain = 0,
                storm = 0,
                snow = 0
            }
            dryingStateData.dryingFactors = {
                night = THMoistureSystem.DEFAULT_DRYING_FACTORS.NIGHT,
                cloud = THMoistureSystem.DEFAULT_DRYING_FACTORS.CLOUD,
                rain = THMoistureSystem.DEFAULT_DRYING_FACTORS.RAIN,
                storm = THMoistureSystem.DEFAULT_DRYING_FACTORS.STORM,
                snow = THMoistureSystem.DEFAULT_DRYING_FACTORS.SNOW,
                wind = THMoistureSystem.DEFAULT_DRYING_FACTORS.WIND,
                humidity = {},
                soilTypes = {}
            }
            local humidityFactors = dryingStateData.dryingFactors.humidity
            local absMinHumidity = self.dryingConditions.absMinHumidity
            local absMaxHumidity = self.dryingConditions.absMaxHumidity
            table.insert(humidityFactors, { value = absMinHumidity, factor = 1 })
            table.insert(humidityFactors, { value = absMaxHumidity, factor = 0 })
            dryingStateData.hasResetPeriods = true
            dryingStateData.resetPeriods = {}
            return dryingStateData
        end
    end
end
function THMoistureSystem:resetFruitTypeDryingState(fruitType, dryingStateIndex, noEventSend)
    local weatherSystem = self.weatherSystem
    local windUpdater = self.windUpdater
    if self.isServer or noEventSend == true then
        local dryingStateData = self:getFruitTypeDryingState(fruitType, dryingStateIndex)
        if dryingStateData == nil then
            return false
        end
        local minTemp, maxTemp = weatherSystem:getCurrentMinMaxTemperatures()
        local _, _, windSpeed = windUpdater:getCurrentValues()
        dryingStateData.averageTemp = (minTemp + maxTemp) / 2
        dryingStateData.averageWind = windSpeed
        local dryingTimes = dryingStateData.dryingTimes
        dryingTimes.total = 0
        dryingTimes.night = 0
        dryingTimes.cloud = 0
        dryingTimes.rain = 0
        dryingTimes.storm = 0
        dryingTimes.snow = 0
        if self.isServer and noEventSend ~= true then
            self:raiseDirtyFlags(self.dirtyFlag)
        end
        return true
    end
    return false
end
function THMoistureSystem:updateFruitTypeDryingStates(elapsedTime)
    local updateSuccess = true
    if self.isServer then
        if not self.isDryingStateUpdateSafe then
            self.dryingStateCachedTime = self.dryingStateCachedTime + elapsedTime
        else
            elapsedTime = elapsedTime + self.dryingStateCachedTime
            self.dryingStateCachedTime = 0
            self.isDryingStateUpdateSafe = false
            local numMoistureFruitTypes = #self.fruitTypes
            if numMoistureFruitTypes > 0 then
                for fruitTypeIndex = 1, numMoistureFruitTypes do
                    local fruitTypeMoistureData = self.fruitTypes[fruitTypeIndex]
                    local numDryingStates = #fruitTypeMoistureData.dryingStates
                    if numDryingStates > 0 then
                        for dryingStateIndex = 1, numDryingStates do
                            if not self:updateFruitTypeDryingState(fruitTypeIndex, dryingStateIndex, elapsedTime) then
                                updateSuccess = false
                            end
                        end
                    end
                end
            end
        end
        self:raiseActive()
    else
        updateSuccess = false
    end
    return updateSuccess
end
function THMoistureSystem:updateFruitTypeDryingState(fruitType, dryingState, elapsedTime, noEventSend)
    local thFruitTypeManager = self.thFruitTypeManager
    local environment = self.environment
    local weatherSystem = self.weatherSystem
    local windUpdater = self.windUpdater
    local updateSuccess = true
    if self.isServer then
        local dryingConditions = self.dryingConditions
        local dryingStateData = self:getFruitTypeDryingState(fruitType, dryingState)
        if dryingStateData == nil then
            local fruitTypeInfo = thFruitTypeManager:getFruitType(fruitType)
            if fruitTypeInfo == nil then
                THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "fruitType", fruitType)
                updateSuccess = false
            elseif THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.ERROR) then
                THUtils.warningMsg("No drying state information found for fruitType [%s] (%s)", fruitTypeInfo.index, fruitTypeInfo.name)
            end
        else
            local fruitTypeMoistureData = dryingStateData.parent
            local dryingTimes = dryingStateData.dryingTimes
            local dryingFactors = dryingStateData.dryingFactors
            local weatherType = weatherSystem:getCurrentWeatherType()
            local isDayTime = environment.isSunOn
            local curTemp = weatherSystem:getCurrentTemperature()
            local curHumidity = dryingConditions.humidity
            local _, _, windSpeed = windUpdater:getCurrentValues()
            local windFactor = MathUtil.clamp(windSpeed / WindUpdater.MAX_SPEED, 0, 1) * dryingFactors.wind
            local humidityFactor = self:getDryingStateHumidityFactor(dryingStateData, curHumidity)
            local weatherFactor = 1
            if weatherType == WeatherType.CLOUDY then
                weatherFactor = dryingFactors.cloud
            elseif weatherType == WeatherType.RAIN then
                weatherFactor = dryingFactors.rain
            elseif weatherType == WeatherType.SNOW then
                weatherFactor = dryingFactors.snow
            end
            if not isDayTime then
                weatherFactor = math.min(weatherFactor, dryingFactors.night)
            end
            local adjustedHumidityFactor = (humidityFactor + windFactor) / 2
            local totalDryingFactor = math.min(adjustedHumidityFactor, weatherFactor)
            local adjustedTime = elapsedTime * totalDryingFactor -- can be negative
            dryingTimes.total = math.max(0, dryingTimes.total + adjustedTime)
            if not isDayTime then
                dryingTimes.night = math.max(0, dryingTimes.night + adjustedTime)
            end
            if weatherType == WeatherType.CLOUDY then
                dryingTimes.cloud = math.max(0, dryingTimes.cloud + adjustedTime)
            elseif weatherType == WeatherType.RAIN then
                dryingTimes.rain = math.max(0, dryingTimes.rain + adjustedTime)
            elseif weatherType == WeatherType.SNOW then
                dryingTimes.snow = math.max(0, dryingTimes.snow + adjustedTime)
            end
            dryingStateData.averageTemp = (dryingStateData.averageTemp + curTemp) / 2
            dryingStateData.averageWind = (dryingStateData.averageWind + windSpeed) / 2
            if THDebugUtil.getIsEnabled(debugFlagId, THDebugLevel.UPDATE) then
                local fruitTypeName = fruitTypeMoistureData.fruitType.name
                local fruitTypeIndex = fruitTypeMoistureData.fruitType.index
                THUtils.displayMsg("Updated drying state [%s] for fruitType [%s] (%s)", dryingStateData.index, fruitTypeIndex, fruitTypeName)
                THUtils.displayMsg("- Wind speed: %s", windSpeed)
                THUtils.displayMsg("- Wind factor: %s", windFactor)
                THUtils.displayMsg("- Humidity: %s", curHumidity)
                THUtils.displayMsg("- Humidity factor: %s", humidityFactor)
                THUtils.displayMsg("- Weather factor: %s", weatherFactor)
                THUtils.displayMsg("- Total drying factor: %s", totalDryingFactor)
                THUtils.printTable(dryingStateData, 1)
            end
            if noEventSend ~= true then
                self:raiseDirtyFlags(self.dirtyFlag)
            end
        end
    else
        updateSuccess = false
    end
    return updateSuccess
end
function THMoistureSystem:getFruitTypeMoisture(fruitType, growthStateId, harvestMultiplier, farmlandId, soilType)
    local pfManager = self.thPFManager
    local dryingConditions = self.dryingConditions
    local dryingStateData, growthState = self:getFruitTypeDryingStateByGrowthState(fruitType, growthStateId)
    local currentMoisture = -1
    local totalDryingFactor = -1
    if dryingStateData ~= nil and growthState >= 0 then
        local fruitTypeMoistureData = dryingStateData.parent
        local fruitTypeData = fruitTypeMoistureData.fruitType.data
        local startMoisture = dryingStateData.startPercent
        local endMoisture = dryingStateData.endPercent
        if not dryingConditions.isInitialized then
            if fruitTypeData ~= nil and growthState ~= nil and growthState >= 0 then
                local minHarvestState = fruitTypeData.minHarvestState or -1
                local maxHarvestState = fruitTypeData.maxHarvestState or -1
                if maxHarvestState > 0 and growthState > maxHarvestState then
                    totalDryingFactor = 1
                elseif minHarvestState > 0 then
                    totalDryingFactor = 0
                end
            end
        else
            local timeToDry = dryingStateData.timeToDry * self.periodLength
            local totalDryingTime = dryingStateData.dryingTimes.total
            if timeToDry <= 0 then
                if totalDryingTime > 0 then
                    totalDryingFactor = 1
                else
                    totalDryingFactor = 0
                end
            else
                local yieldDryingBonus = 0
                local farmlandDryingFactor = 1
                local soilTypeDryingFactor = 1
                if harvestMultiplier ~= nil then
                    yieldDryingBonus = THUtils.getFactorInRange(1, 3, harvestMultiplier)
                end
                if farmlandId ~= nil then
                    farmlandDryingFactor = self:getFarmlandDryingFactor(farmlandId)
                end
                if soilType ~= nil then
                    local soilTypeFactor = dryingStateData.dryingFactors.soilTypes[soilType]
                    if soilTypeFactor ~= nil then
                        soilTypeDryingFactor = soilTypeFactor
                    else
                        local soilTypeInfo = pfManager:getSoilType(soilType)
                        if soilTypeInfo ~= nil then
                            local minPotential = pfManager.minSoilTypeYieldPotential
                            local maxPotential = pfManager.maxSoilTypeYieldPotential
                            if minPotential ~= nil and maxPotential ~= nil then
                                local curPotential = soilTypeInfo.yieldPotential or maxPotential
                                soilTypeDryingFactor = THUtils.getFactorInRange(minPotential, maxPotential, curPotential)
                            end
                        end
                    end
                end
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local fruitTypeIndex = fruitTypeMoistureData.fruitType.index
                    local fruitTypeName = fruitTypeMoistureData.fruitType.name
                    THUtils.displayMsg("Current moisture for fruitType: [%s] (%s)", fruitTypeIndex, fruitTypeName)
                    THUtils.displayMsg("- Drying time: %s", totalDryingTime)
                    THUtils.displayMsg("- Time to dry: %s", timeToDry)
                    THUtils.displayMsg("- Yield drying bonus: %s", yieldDryingBonus)
                    THUtils.displayMsg("- Farmland drying factor: %s", farmlandDryingFactor)
                    THUtils.displayMsg("- SoilType drying factor: %s", soilTypeDryingFactor)
                end
                totalDryingTime = totalDryingTime + (totalDryingTime * yieldDryingBonus)
                totalDryingTime = (totalDryingTime * 0.5) + (totalDryingTime * 0.5 * farmlandDryingFactor)
                totalDryingTime = (totalDryingTime * 0.5) + (totalDryingTime * 0.5 * soilTypeDryingFactor)
                if totalDryingTime < timeToDry then
                    totalDryingFactor = MathUtil.clamp(totalDryingTime / timeToDry, 0, 1)
                else
                    totalDryingFactor = 1
                end
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    THUtils.displayMsg("- Drying time (adjusted): %s", totalDryingTime)
                    THUtils.displayMsg("- Total drying factor: %s", totalDryingFactor)
                    THUtils.displayMsg("")
                end
            end
        end
        if totalDryingFactor >= 0 then
            currentMoisture = startMoisture + ((endMoisture - startMoisture) * totalDryingFactor)
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THUtils.displayMsg("- Starting moisture: %s", startMoisture)
                THUtils.displayMsg("- Ending moisture: %s", endMoisture)
                THUtils.displayMsg("- Current moisture: %s", currentMoisture)
                THUtils.displayMsg("")
            end
        end
    end
    return currentMoisture, totalDryingFactor
end
function THMoistureSystem:getDryingStateHumidityFactor(dryingStateData, humidity)
    local humidityFactors = dryingStateData.dryingFactors.humidity
    local numHumidityFactors = #humidityFactors
    local humidityFactor = 1
    if numHumidityFactors > 0 then
        for factorIndex = 1, numHumidityFactors do
            local factorEntry = humidityFactors[factorIndex]
            if humidity <= factorEntry.value then
                if factorIndex == 1 then
                    humidityFactor = factorEntry.factor
                else
                    local lastFactorEntry = humidityFactors[factorIndex - 1]
                    local humidityAlpha = THUtils.getFactorInRange(lastFactorEntry.value, factorEntry.value, humidity)
                    humidityFactor = THUtils.getValueInRange(lastFactorEntry.factor, factorEntry.factor, humidityAlpha)
                end
                break
            end
        end
    end
    return humidityFactor
end
function THMoistureSystem:setFillTypeMoistureFactors(fillType, moisture, priceFactor, yieldFactor, decayFactor)
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType, true) -- force
    if fillTypeMoistureData == nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "fillType", fillType)
    elseif THUtils.argIsValid(type(moisture) == "number" and moisture >= 0, "moisture", moisture)
        and THUtils.argIsValid(type(priceFactor) == "number" and priceFactor >= 0, "priceFactor", priceFactor)
        and THUtils.argIsValid(type(yieldFactor) == "number" and yieldFactor >= 0, "yieldFactor", yieldFactor)
        and THUtils.argIsValid(type(decayFactor) == "number" and decayFactor >= 0 and decayFactor <= 1, "decayFactor", decayFactor)
        and fillTypeMoistureData.isEnabled
    then
        local curveEntries = fillTypeMoistureData.factors.curve
        local factorData = nil
        for _, otherFactorData in pairs(curveEntries) do
            if moisture == otherFactorData.moisture then
                factorData = otherFactorData
                break
            end
        end
        if factorData == nil then
            factorData = {
                moisture = moisture,
                priceFactor = priceFactor,
                yieldFactor = yieldFactor,
                decayFactor = decayFactor
            }
            table.insert(curveEntries, factorData)
            table.sort(curveEntries, function(pa, pb)
                return pa.moisture < pb.moisture
            end)
        else
            factorData.priceFactor = priceFactor
            factorData.yieldFactor = yieldFactor
            factorData.decayFactor = decayFactor
        end
        return true
    end
    return false
end
function THMoistureSystem:getFillTypeMoistureData(fillType, force)
    local thFillTypeManager = self.thFillTypeManager
    local fillTypeInfo = thFillTypeManager:getFillType(fillType)
    if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
        local fillTypeMoistureData = self.fillTypes[fillTypeInfo.index]
        if fillTypeMoistureData ~= nil then
            if (fillTypeMoistureData.isEnabled and fillTypeMoistureData.hasMoisture)
                or force == true
            then
                return fillTypeMoistureData
            end
        end
    end
end
function THMoistureSystem:getFillTypeMoisture(fruitType, fillType, growthStateId, harvestMultiplier, farmlandId, soilType)
    local fruitTypeMoisture = self:getFruitTypeMoisture(fruitType, growthStateId, harvestMultiplier, farmlandId, soilType)
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType)
    local fillTypeMoisture = -1
    if fillTypeMoistureData ~= nil and fruitTypeMoisture >= 0 then
        local minMoisture = fillTypeMoistureData.minPercent
        local maxMoisture = fillTypeMoistureData.maxPercent
        fillTypeMoisture = MathUtil.clamp(fruitTypeMoisture, minMoisture, maxMoisture)
    end
    return fillTypeMoisture
end
function THMoistureSystem:getFillTypeMoistureFactors(fillType, moisture, useSettings)
    local fillTypeMoistureData = self:getFillTypeMoistureData(fillType)
    local priceFactor = 1
    local yieldFactor = 1
    local decayFactor = 0
    if fillTypeMoistureData ~= nil then
        local moistureFactors = fillTypeMoistureData.factors
        local factorsCurve = moistureFactors.curve
        local numCurveEntries = #factorsCurve
        if numCurveEntries > 0 then
            moisture = MathUtil.clamp(moisture, fillTypeMoistureData.minPercent, fillTypeMoistureData.maxPercent)
            for curveIndex = 1, numCurveEntries do
                local curveEntry = factorsCurve[curveIndex]
                if moisture <= curveEntry.moisture then
                    local lastCurveEntry = factorsCurve[curveIndex - 1]
                    if lastCurveEntry == nil then
                        priceFactor = curveEntry.priceFactor
                        yieldFactor = curveEntry.yieldFactor
                        decayFactor = curveEntry.decayFactor
                    else
                        local moistureAlpha = THUtils.getFactorInRange(lastCurveEntry.moisture, curveEntry.moisture, moisture)
                        priceFactor = THUtils.getValueInRange(lastCurveEntry.priceFactor, curveEntry.priceFactor, moistureAlpha)
                        yieldFactor = THUtils.getValueInRange(lastCurveEntry.yieldFactor, curveEntry.yieldFactor, moistureAlpha)
                        decayFactor = THUtils.getValueInRange(lastCurveEntry.decayFactor, curveEntry.decayFactor, moistureAlpha)
                    end
                    break
                end
            end
        end
    end
    if priceFactor ~= nil and priceFactor ~= 1 then
        if useSettings ~= false then
            local deltaPriceFactor = 0
            if priceFactor > 1 then
                local priceBonusFactor = self.thSettingsManager:getSettingValue(THMoistureSystem.SETTING.PRICE_BONUS_FACTOR) or 1
                deltaPriceFactor = priceFactor - 1
                priceFactor = 1 + (deltaPriceFactor * priceBonusFactor)
            else
                local pricePenaltyFactor = self.thSettingsManager:getSettingValue(THMoistureSystem.SETTING.PRICE_PENALTY_FACTOR) or 1
                deltaPriceFactor = 1 - priceFactor
                priceFactor = 1 - (deltaPriceFactor * pricePenaltyFactor)
            end
        end
        priceFactor = math.max(0.000001, priceFactor)
    end
    if yieldFactor ~= nil and yieldFactor ~= 1 then
        if useSettings ~= false then
            local deltaYieldFactor = 0
            if yieldFactor > 1 then
                local yieldBonusFactor = self.thSettingsManager:getSettingValue(THMoistureSystem.SETTING.YIELD_BONUS_FACTOR) or 1
                deltaYieldFactor = yieldFactor - 1
                yieldFactor = 1 + (deltaYieldFactor * yieldBonusFactor)
            else
                local yieldPenaltyFactor = self.thSettingsManager:getSettingValue(THMoistureSystem.SETTING.YIELD_PENALTY_FACTOR) or 1
                deltaYieldFactor = 1 - yieldFactor
                yieldFactor = 1 - (deltaYieldFactor * yieldPenaltyFactor)
            end
        end
        yieldFactor = math.max(0.000001, yieldFactor)
    end
    if decayFactor ~= nil then
        if useSettings ~= false then
            local globalDecayFactor = self.thSettingsManager:getSettingValue(THMoistureSystem.SETTING.STORAGE_LOSS_FACTOR) or 1
            decayFactor = decayFactor * globalDecayFactor
        end
        decayFactor = MathUtil.clamp(decayFactor, 0, 1)
    end
    return priceFactor, yieldFactor, decayFactor
end
THMoistureUtil = {}
function THMoistureUtil.calculateDewPoint(temp, humidity)
    local vA = 17.625
    local vB = 243.04
    local vH = math.log(humidity) + ((vA * temp) / (vB + temp))
    return math.min(temp, (vB * vH) / (vA - vH))
end
function THMoistureUtil.calculateHumidity(temp, dewPoint)
    local vA = 17.625
    local vB = 243.04
    local vD = math.exp((vA * dewPoint) / (vB + dewPoint))
    local vT = math.exp((vA * temp) / (vB + temp))
    return MathUtil.clamp(vD / vT, 0, 1)
end