-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thMain = g_thMoistureSystem
THMoistureLoadingStation = {}
local THMoistureLoadingStation_mt = THUtils.createClass(THMoistureLoadingStation)
local debugFlagId = THDebugUtil.createDebugFlag("THMoistureLoadingStation")
THMoistureLoadingStation.debugFlagId = debugFlagId
local function getCustomData(self)
    return thMain:getDataTable(self)
end
THMoistureLoadingStation.getCustomData = getCustomData
function THMoistureLoadingStation.newData(parent, rootEnv, customMt)
    customMt = customMt or THMoistureLoadingStation_mt
    local self = setmetatable({}, customMt)
    self.rootEnvironment = rootEnv
    self.parent = parent
    self.moistureFillLevels = {}
    self.moistureFillLevelMapping = {}
    self.areMoistureFillLevelsDirty = true
    return self
end
function THMoistureLoadingStation:setMoistureDirty()
    self.areMoistureFillLevelsDirty = true
end
function THMoistureLoadingStation:updateMoistureFillLevels(force)
    local rootEnv = self.rootEnvironment
    local station = self.parent
    if self.areMoistureFillLevelsDirty or force == true then
        local sourceStorages = station.sourceStorages
        local moistureFillLevels = THUtils.clearTable(self.moistureFillLevels)
        local moistureFillLevelMapping = THUtils.clearTable(self.moistureFillLevelMapping)
        if sourceStorages ~= nil then
            for _, storage in pairs(sourceStorages) do
                local storageData = THMoistureStorage.getCustomData(storage)
                local storageFillLevels = storage:getFillLevels()
                if storageData ~= nil and storageFillLevels ~= nil then
                    for fillTypeIndex, fillLevel in pairs(storageFillLevels) do
                        if fillLevel > 0 and fillTypeIndex ~= FillType.UNKNOWN then
                            local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillTypeIndex)
                            local fillTypeMoisture = storageData:getFillTypeMoisture(fillTypeIndex)
                            if fillTypeMoistureData ~= nil and fillTypeMoisture >= 0 then
                                local fillLevelIndex = moistureFillLevelMapping[fillTypeIndex] or 0
                                local fillLevelData = moistureFillLevels[fillLevelIndex]
                                if fillLevelIndex <= 0 or fillLevelData == nil then
                                    fillLevelIndex = #moistureFillLevels + 1
                                    fillLevelData = {
                                        index = fillLevelIndex,
                                        fillTypeIndex = fillTypeIndex,
                                        fillTypeTitle = fillTypeMoistureData.fillType.title,
                                        fillLevel = fillLevel,
                                        moisture = fillTypeMoisture
                                    }
                                    moistureFillLevels[fillLevelIndex] = fillLevelData
                                    moistureFillLevelMapping[fillTypeIndex] = fillLevelIndex
                                else
                                    local totalFillLevel = fillLevelData.fillLevel + fillLevel
                                    local minMoisture = fillTypeMoistureData.minPercent
                                    local maxMoisture = fillTypeMoistureData.maxPercent
                                    local fillPercent = fillLevel / totalFillLevel
                                    fillLevelData.fillLevel = totalFillLevel
                                    fillLevelData.moisture = MathUtil.clamp((fillTypeMoisture * fillPercent) + (fillLevelData.moisture * (1 - fillPercent)), minMoisture, maxMoisture)
                                end
                            end
                        end
                    end
                end
            end
        end
        self.areMoistureFillLevelsDirty = false
    end
end
function THMoistureLoadingStation:getFillTypeMoisture(fillType)
    local rootEnv = self.rootEnvironment
    local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillType)
    if fillTypeMoistureData ~= nil then
        self:updateMoistureFillLevels()
        local fillLevelIndex = self.moistureFillLevelMapping[fillTypeMoistureData.fillType.index] or 0
        local fillLevelData = self.moistureFillLevels[fillLevelIndex]
        if fillLevelIndex > 0 and fillLevelData ~= nil then
            return fillLevelData.moisture or -1
        end
    end
    return -1
end
function THMoistureLoadingStation.hook_new(superFunc, ...)
    local function appendFunc(rStation, ...)
        if rStation ~= nil then
            THUtils.call(function()
                local customData = thMain:createDataTable(rStation, false, THMoistureLoadingStation)
            end)
        end
        return rStation, ...
    end
    return appendFunc(superFunc(...))
end
function THMoistureLoadingStation:hook_addFillLevelToFillableObject(superFunc, object, fillUnitIndex, fillTypeIndex, deltaFill, ...)
    local customData = getCustomData(self)
    if customData ~= nil then
        THUtils.call(function()
            if object ~= nil and object.isa ~= nil
                and fillTypeIndex ~= nil and fillTypeIndex ~= FillType.UNKNOWN
            then
                local fillTypeMoisture = customData:getFillTypeMoisture(fillTypeIndex)
                if fillTypeMoisture >= 0 then
                    if object:isa(Vehicle) and fillUnitIndex ~= nil
                        and object.thAA_setFillUnitInputMoisture ~= nil
                    then
                        object:thAA_setFillUnitInputMoisture(fillUnitIndex, fillTypeIndex, fillTypeMoisture)
                    elseif object:isa(UnloadTrigger) and object.target ~= nil then
                        local objectData = THMoistureUnloadingStation.getCustomData(object.target)
                        if objectData ~= nil and objectData.setFillTypeInputMoisture ~= nil then
                            objectData:setFillTypeInputMoisture(fillTypeIndex, fillTypeMoisture)
                        end
                    end
                end
            end
        end)
    end
    return superFunc(self, object, fillUnitIndex, fillTypeIndex, deltaFill, ...)
end
thMain:addDelayedLoadTask(function()
    THUtils.setFunctionHook("LoadingStation", "new", THMoistureLoadingStation, nil, false)
    THUtils.setFunctionHook("LoadingStation", "addFillLevelToFillableObject", THMoistureLoadingStation)
end)