-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thMain = g_thMoistureSystem
local thModName = thMain.modName
THMoistureStorage = {}
local THMoistureStorage_mt = THUtils.createClass(THMoistureStorage)
local debugFlagId = THDebugUtil.createDebugFlag("THMoistureStorage")
THMoistureStorage.debugFlagId = debugFlagId
local function getCustomData(self)
    return thMain:getDataTable(self)
end
THMoistureStorage.getCustomData = getCustomData
function THMoistureStorage.newData(parent, rootEnv, customMt)
    customMt = customMt or THMoistureStorage_mt
    local self = setmetatable({}, customMt)
    self.rootEnvironment = rootEnv
    self.parent = parent
    self.isClient = parent.isClient
    self.isServer = parent.isServer
    self.fillTypeMoisture = {}
    self.moistureChangedListeners = {}
    self.fillTypeInputMoisture = {}
    self.dirtyFlag = parent:getNextDirtyFlag()
    local function onFillLevelChanged(pFillTypeIndex, pAppliedDelta, ...)
        return self:onFillLevelChanged(pFillTypeIndex, pAppliedDelta, ...)
    end
    parent:addFillLevelChangedListeners(onFillLevelChanged)
    return self
end
function THMoistureStorage:onFillLevelChanged(fillTypeIndex, appliedDelta, ...)
    self:updateFillTypeMoisture(fillTypeIndex, nil, appliedDelta)
end
function THMoistureStorage:readStorageFromStream(streamId, connection)
    local storage = self.parent
    local sortedFillTypes = storage.sortedFillTypes
    if sortedFillTypes ~= nil then
        local numFillTypes = #sortedFillTypes
        if numFillTypes > 0 then
            for i = 1, numFillTypes do
                local fillTypeIndex = sortedFillTypes[i]
                self.fillTypeMoisture[fillTypeIndex] = streamReadFloat32(streamId)
            end
        end
    end
end
function THMoistureStorage:writeStorageToStream(streamId, connection)
    local storage = self.parent
    local sortedFillTypes = storage.sortedFillTypes
    if sortedFillTypes ~= nil then
        local numFillTypes = #sortedFillTypes
        if numFillTypes > 0 then
            for i = 1, numFillTypes do
                local fillTypeIndex = sortedFillTypes[i]
                streamWriteFloat32(streamId, self.fillTypeMoisture[fillTypeIndex] or -1)
            end
        end
    end
end
function THMoistureStorage:setMoistureDirty()
    local storage = self.parent
    if storage ~= nil then
        local thMoistureStation = nil
        if storage.loadingStations ~= nil then
            for _, loadingStation in pairs(storage.loadingStations) do
                thMoistureStation = THMoistureLoadingStation.getCustomData(loadingStation)
                if thMoistureStation ~= nil and thMoistureStation.setMoistureDirty ~= nil then
                    thMoistureStation:setMoistureDirty()
                end
            end
        end
        if storage.unloadingStations ~= nil then
            for _, unloadingStation in pairs(storage.unloadingStations) do
                thMoistureStation = THMoistureUnloadingStation.getCustomData(unloadingStation)
                if thMoistureStation ~= nil and thMoistureStation.setMoistureDirty ~= nil then
                    thMoistureStation:setMoistureDirty()
                end
            end
        end
    end
end
function THMoistureStorage:getFillTypeMoisture(fillType)
    local rootEnv = self.rootEnvironment
    local storage = self.parent
    local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillType)
    if fillTypeMoistureData ~= nil then
        local fillTypeIndex = fillTypeMoistureData.fillType.index
        local storageFillLevel = storage:getFillLevel(fillTypeIndex) or 0
        if storageFillLevel > 0 then
            local storageMoisture = self.fillTypeMoisture[fillTypeIndex] or -1
            if storageMoisture < 0 then
                storageMoisture = fillTypeMoistureData.defPercent
            else
                storageMoisture = MathUtil.clamp(storageMoisture, fillTypeMoistureData.minPercent, fillTypeMoistureData.maxPercent)
            end
            storageMoisture = THUtils.round(storageMoisture, 4)
            storageFillLevel = THUtils.round(storageFillLevel, 4)
            return storageMoisture, storageFillLevel
        end
    end
    return -1, 0
end
function THMoistureStorage:setFillTypeMoisture(fillType, moisture, noEventSend)
    if THUtils.argIsValid(type(moisture) == "number", "moisture", moisture) then
        local rootEnv = self.rootEnvironment
        local storage = self.parent
        local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillType)
        if fillTypeMoistureData ~= nil then
            local fillTypeIndex = fillTypeMoistureData.fillType.index
            local fillTypeName = fillTypeMoistureData.fillType.name
            local storageFillLevel = storage:getFillLevel(fillTypeIndex) or 0
            local oldMoisture = self.fillTypeMoisture[fillTypeIndex] or -1
            local newMoisture = oldMoisture
            if storageFillLevel > 0 then
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                local defMoisture = fillTypeMoistureData.defPercent
                if moisture < 0 then
                    newMoisture = defMoisture
                else
                    newMoisture = MathUtil.clamp(moisture, minMoisture, maxMoisture)
                end
            else
                newMoisture = -1
            end
            if math.abs(newMoisture - oldMoisture) > 0.000001 then
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    THDebugUtil.printMsg(debugFlagId, nil, "Moisture values for fillType: %s", fillTypeName)
                    THDebugUtil.printMsg(debugFlagId, nil, "- Old moisture: %s", oldMoisture)
                    THDebugUtil.printMsg(debugFlagId, nil, "- New moisture: %s", newMoisture)
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                end
                self.fillTypeMoisture[fillTypeIndex] = newMoisture
                for _, onChangeFunc in pairs(self.moistureChangedListeners) do
                    THUtils.call(onChangeFunc, storage, newMoisture, oldMoisture)
                end
                self:setMoistureDirty()
                if self.isServer and noEventSend ~= true then
                    storage:raiseDirtyFlags(self.dirtyFlag)
                end
            end
        end
    end
end
function THMoistureStorage:updateFillTypeMoisture(fillType, inputMoisture, inputDelta, noEventSend)
    if self.isServer then
        local rootEnv = self.rootEnvironment
        local storage = self.parent
        local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillType)
        if fillTypeMoistureData ~= nil then
            local fillTypeIndex = fillTypeMoistureData.fillType.index
            local fillTypeName = fillTypeMoistureData.fillType.name
            local storageFillLevel = storage:getFillLevel(fillTypeIndex) or 0
            local oldMoisture = THUtils.getNoNil(self.fillTypeMoisture[fillTypeIndex], -1)
            local newMoisture = oldMoisture
            if storageFillLevel <= 0 then
                newMoisture = -1
            else
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                local defMoisture = fillTypeMoistureData.defPercent
                if inputDelta == nil then
                    inputDelta = 0
                else
                    inputDelta = MathUtil.clamp(inputDelta, -storageFillLevel, storageFillLevel)
                end
                if inputMoisture == nil then
                    local inputMoistureData = self.fillTypeInputMoisture[fillTypeIndex]
                    if inputMoistureData ~= nil and inputMoistureData.moisture ~= nil
                        and inputMoistureData.moisture >= 0
                    then
                        inputMoisture = inputMoistureData.moisture
                    end
                end
                if inputMoisture ~= nil then
                    if inputMoisture < 0 then
                        inputMoisture = defMoisture
                    else
                        inputMoisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                    end
                end
                if oldMoisture < 0 then
                    newMoisture = inputMoisture or defMoisture
                elseif inputDelta > 0 and inputMoisture ~= nil then
                    local fillPercent = inputDelta / storageFillLevel
                    newMoisture = MathUtil.clamp((inputMoisture * fillPercent) + (oldMoisture * (1 - fillPercent)), minMoisture, maxMoisture)
                else
                    newMoisture = MathUtil.clamp(oldMoisture, minMoisture, maxMoisture)
                end
            end
            if math.abs(newMoisture - oldMoisture) > 0.000001 then
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    THDebugUtil.printMsg(debugFlagId, nil, "Moisture values for fillType: %s", fillTypeName)
                    THDebugUtil.printMsg(debugFlagId, nil, "- Old moisture: %s", oldMoisture)
                    THDebugUtil.printMsg(debugFlagId, nil, "- New moisture: %s", newMoisture)
                    if inputMoisture ~= nil then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Input moisture: %s", inputMoisture)
                    end
                    if inputDelta ~= nil then
                        THDebugUtil.printMsg(debugFlagId, nil, "- Input delta: %s", inputDelta)
                    end
                    THDebugUtil.printMsg(debugFlagId, nil, "")
                end
                self.fillTypeMoisture[fillTypeIndex] = newMoisture
                for _, onChangeFunc in pairs(self.moistureChangedListeners) do
                    THUtils.call(onChangeFunc, storage, newMoisture, oldMoisture)
                end
                self:setMoistureDirty()
                if noEventSend ~= true then
                    storage:raiseDirtyFlags(self.dirtyFlag)
                end
            end
        end
    end
end
function THMoistureStorage:setFillTypeInputMoisture(fillType, inputMoisture)
    if THUtils.argIsValid(type(inputMoisture) == "number", "inputMoisture", inputMoisture) then
        local rootEnv = self.rootEnvironment
        local storage = self.parent
        local thFillTypeManager = rootEnv.thFillTypeManager
        local fillTypeInfo = thFillTypeManager:getFillType(fillType)
        if fillTypeInfo ~= nil and fillTypeInfo.index ~= nil then
            local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillTypeInfo.index)
            if fillTypeMoistureData == nil or inputMoisture < 0 then
                THUtils.clearTable(self.fillTypeInputMoisture[fillTypeInfo.index])
            else
                local inputMoistureData = THUtils.createSubTable(self.fillTypeInputMoisture, fillTypeInfo.index, true)
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                inputMoistureData.fillTypeIndex = fillTypeInfo.index
                inputMoistureData.moisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                inputMoistureData.resetTime = 50
                THDebugUtil.printMsg(debugFlagId, nil, "Input moisture set for fillType: [%s] (%s)", fillTypeInfo.index, fillTypeInfo.name)
                THDebugUtil.printTable(debugFlagId, nil, inputMoistureData)
                storage:raiseActive()
            end
        end
    end
end
function THMoistureStorage.hook_registerSavegameXMLPaths(superFunc, xmlSchema, xmlPath, ...)
    local function appendFunc(...)
        THUtils.call(function()
            local xmlBasePath = string.format("%s.%s", xmlPath, thModName)
            THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlBasePath .. ".fillType(?)", "#name", "Name of moisture fillType")
            THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlBasePath .. ".fillType(?)", "#moisture", "Current fillType moisture percentage")
        end)
        return ...
    end
    return appendFunc(superFunc(xmlSchema, xmlPath))
end
function THMoistureStorage.hook_new(superFunc, ...)
    local function appendFunc(rStorage, ...)
        if rStorage ~= nil then
            THUtils.call(function()
                local customData = thMain:createDataTable(rStorage, false, THMoistureStorage)
            end)
        end
        return rStorage, ...
    end
    return appendFunc(superFunc(...))
end
function THMoistureStorage:hook_update(superFunc, dt, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                local hasInputMoisture = false
                for _, inputMoistureData in pairs(customData.fillTypeInputMoisture) do
                    if inputMoistureData.resetTime ~= nil and inputMoistureData.resetTime > 0 then
                        inputMoistureData.resetTime = inputMoistureData.resetTime - dt
                        if inputMoistureData.resetTime <= 0 then
                            THUtils.clearTable(inputMoistureData)
                        else
                            hasInputMoisture = true
                        end
                    elseif next(inputMoistureData) ~= nil then
                        THUtils.clearTable(inputMoistureData)
                    end
                end
                if hasInputMoisture then
                    self:raiseActive()
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THMoistureStorage:hook_readStream(superFunc, streamId, connection, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and connection:getIsServer() then
            customData:readStorageFromStream(streamId, connection)
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, ...))
end
function THMoistureStorage:hook_writeStream(superFunc, streamId, connection, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and not connection:getIsServer() then
            customData:writeStorageToStream(streamId, connection)
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, ...))
end
function THMoistureStorage:hook_readUpdateStream(superFunc, streamId, timeStamp, connection, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and connection:getIsServer() then
            if streamReadBool(streamId) then -- dirtyMask
                customData:readStorageFromStream(streamId, connection)
            end
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, timeStamp, connection, ...))
end
function THMoistureStorage:hook_writeUpdateStream(superFunc, streamId, connection, dirtyMask, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and not connection:getIsServer() then
            if streamWriteBool(streamId, bitAND(dirtyMask, customData.dirtyFlag) ~= 0) then
                customData:writeStorageToStream(streamId, connection)
            end
        end
        return ...
    end
    return appendFunc(superFunc(self, streamId, connection, dirtyMask, ...))
end
function THMoistureStorage:hook_loadFromXMLFile(superFunc, xmlFile, xmlKey, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        THUtils.call(function()
            if self.isServer and customData ~= nil then
                local rootEnv = customData.rootEnvironment
                local thFillTypeManager = rootEnv.thFillTypeManager
                local xmlBaseKey = string.format("%s.%s", xmlKey, thModName)
                xmlFile:iterate(xmlBaseKey .. ".fillType", function(_, pFillTypeKey)
                    local fillTypeName = THUtils.getXMLValue(xmlFile, pFillTypeKey, "#name")
                    local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
                    if fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN then
                        local fillTypeIndex = fillTypeInfo.index
                        local fillTypeMoisture = THUtils.getXMLValue(xmlFile, pFillTypeKey, "#moisture")
                        if fillTypeMoisture ~= nil then
                            customData:setFillTypeMoisture(fillTypeIndex, fillTypeMoisture, true)
                        end
                    end
                end)
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, xmlFile, xmlKey, ...))
end
function THMoistureStorage:hook_saveToXMLFile(superFunc, xmlFile, xmlKey, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if self.isServer and customData ~= nil then
            THUtils.call(function()
                local rootEnv = customData.rootEnvironment
                local thFillTypeManager = rootEnv.thFillTypeManager
                local xmlBaseKey = string.format("%s.%s", xmlKey, thModName)
                local fillTypeXMLIndex = 0
                for fillTypeIndex, fillTypeMoisture in pairs(customData.fillTypeMoisture) do
                    local fillTypeInfo = thFillTypeManager:getFillType(fillTypeIndex)
                    local storageFillLevel = self:getFillLevel(fillTypeIndex) or 0
                    if storageFillLevel > 0 and fillTypeMoisture >= 0
                        and fillTypeInfo ~= nil and fillTypeInfo.index ~= FillType.UNKNOWN
                    then
                        local fillTypeKey = string.format("%s.fillType(%d)", xmlBaseKey, fillTypeXMLIndex)
                        xmlFile:setString(fillTypeKey .. "#name", fillTypeInfo.name)
                        xmlFile:setFloat(fillTypeKey .. "#moisture", fillTypeMoisture)
                        fillTypeXMLIndex = fillTypeXMLIndex + 1
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, xmlFile, xmlKey, ...))
end
thMain:addDelayedLoadTask(function()
    THUtils.setFunctionHook("Storage", "registerSavegameXMLPaths", THMoistureStorage, nil, false)
    THUtils.setFunctionHook("Storage", "new", THMoistureStorage, nil, false)
    THUtils.setFunctionHook("Storage", "update", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "readStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "writeStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "readUpdateStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "writeUpdateStream", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "loadFromXMLFile", THMoistureStorage)
    THUtils.setFunctionHook("Storage", "saveToXMLFile", THMoistureStorage)
end)