-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thMain = g_thMoistureSystem
THMoistureUnloadingStation = {}
local THMoistureUnloadingStation_mt = THUtils.createClass(THMoistureUnloadingStation, nil, false) -- unprotected/changeable
local debugFlagId = THDebugUtil.createDebugFlag("THMoistureUnloadingStation")
THMoistureUnloadingStation.debugFlagId = debugFlagId
local function getCustomData(self)
    return thMain:getDataTable(self)
end
THMoistureUnloadingStation.getCustomData = getCustomData
THMoistureSellingStation = {}
function THMoistureUnloadingStation.newData(parent, rootEnv, customMt)
    customMt = customMt or THMoistureUnloadingStation_mt
    local self = setmetatable({}, customMt)
    self.rootEnvironment = rootEnv
    self.parent = parent
    self.fillTypeInputMoisture = {} -- used mainly with selling stations
    self.hasMoistureMoneyChange = false
    return self
end
function THMoistureUnloadingStation:setMoistureDirty()
end
function THMoistureUnloadingStation:setFillTypeInputMoisture(fillType, inputMoisture)
    if THUtils.argIsValid(type(inputMoisture) == "number", "inputMoisture", inputMoisture) then
        local rootEnv = self.rootEnvironment
        local station = self.parent
        local thFillTypeManager = rootEnv.thFillTypeManager
        local fillTypeInfo = thFillTypeManager:getFillType(fillType)
        if fillTypeInfo ~= nil and fillTypeInfo.index ~= nil then
            local fillTypeIndex = fillTypeInfo.index
            local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillTypeIndex)
            if fillTypeMoistureData == nil or inputMoisture < 0 then
                THUtils.clearTable(self.fillTypeInputMoisture[fillTypeIndex])
            else
                local minMoisture = fillTypeMoistureData.minPercent
                local maxMoisture = fillTypeMoistureData.maxPercent
                local inputMoistureData = THUtils.createSubTable(self.fillTypeInputMoisture, fillTypeIndex, true)
                inputMoistureData.fillTypeIndex = fillTypeIndex
                inputMoistureData.moisture = MathUtil.clamp(inputMoisture, minMoisture, maxMoisture)
                inputMoistureData.resetTime = 50
                THDebugUtil.printMsg(debugFlagId, nil, "Input moisture set for fillType: [%s] (%s)", fillTypeInfo.index, fillTypeInfo.name)
                THDebugUtil.printTable(debugFlagId, nil, inputMoistureData)
                local targetStorages = station.targetStorages
                if targetStorages ~= nil then
                    for _, storage in pairs(targetStorages) do
                        local storageData = THMoistureStorage.getCustomData(storage)
                        if storageData ~= nil then
                            storageData:setFillTypeInputMoisture(fillType, inputMoisture)
                        end
                    end
                end
                station:raiseActive()
            end
        end
    end
end
function THMoistureUnloadingStation:getFillTypeInputMoisture(fillType)
    local rootEnv = self.rootEnvironment
    local fillTypeMoistureData = rootEnv:getFillTypeMoistureData(fillType)
    if fillTypeMoistureData ~= nil then
        local inputMoistureData = self.fillTypeInputMoisture[fillTypeMoistureData.fillType.index]
        if inputMoistureData ~= nil then
            return inputMoistureData.moisture or -1
        end
    end
    return -1
end
function THMoistureUnloadingStation.hook_new(superFunc, ...)
    local function appendFunc(rStation, ...)
        if rStation ~= nil then
            THUtils.call(function()
                local customData = thMain:createDataTable(rStation, false, THMoistureUnloadingStation)
            end)
        end
        return rStation, ...
    end
    return appendFunc(superFunc(...))
end
function THMoistureSellingStation:hook_update(superFunc, dt, ...)
    local customData = getCustomData(self)
    local doMoneyChange = false
    if customData ~= nil and customData.hasMoistureMoneyChange
        and self.lastMoneyChange ~= nil and self.lastMoneyChange > 0
        and self.lastMoneyChangeFarmId ~= nil
    then
        doMoneyChange = true
    end
    local function appendFunc(...)
        if customData ~= nil and doMoneyChange then
            THUtils.call(function()
                local rootEnv = customData.rootEnvironment
                local mission = rootEnv.mission
                if self.lastMoneyChange <= 0 then
                    local moneyChangeType = rootEnv.moneyChangeTypes.harvestMoisture
                    mission:showMoneyChange(moneyChangeType, nil, false, self.lastMoneyChangeFarmId)
                    customData.hasMoistureMoneyChange = false
                end
                local hasInputMoisture = false
                for _, inputMoistureData in pairs(customData.fillTypeInputMoisture) do
                    if inputMoistureData.resetTime ~= nil and inputMoistureData.resetTime > 0 then
                        inputMoistureData.resetTime = inputMoistureData.resetTime - dt
                        if inputMoistureData.resetTime <= 0 then
                            THUtils.clearTable(inputMoistureData)
                        else
                            hasInputMoisture = true
                        end
                    elseif next(inputMoistureData) ~= nil then
                        THUtils.clearTable(inputMoistureData)
                    end
                end
                if hasInputMoisture then
                    self:raiseActive()
                end
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, dt, ...))
end
function THMoistureSellingStation:hook_sellFillType(superFunc, farmId, deltaFill, fillTypeIndex, toolType, extraAttributes, ...)
    local customData = getCustomData(self)
    local function appendFunc(rPrice, ...)
        if customData ~= nil and rPrice ~= nil and rPrice > 0
            and fillTypeIndex ~= nil and fillTypeIndex ~= FillType.UNKNOWN
        then
            THUtils.call(function()
                local rootEnv = customData.rootEnvironment
                local mission = rootEnv.mission
                local thFillTypeManager = rootEnv.thFillTypeManager
                local fillTypeMoisture = customData:getFillTypeInputMoisture(fillTypeIndex)
                local moneyChangeType = rootEnv.moneyChangeTypes.harvestMoisture
                if fillTypeMoisture >= 0 then
                    local priceFactor = rootEnv:getFillTypeMoistureFactors(fillTypeIndex, fillTypeMoisture)
                    local moistureBonus = (rPrice * priceFactor) - rPrice
                    if math.abs(moistureBonus) > 0.001 then
                        mission:addMoney(moistureBonus, farmId, moneyChangeType, true)
                        customData.hasMoistureMoneyChange = true
                        if THDebugUtil.getIsEnabled(debugFlagId) then
                            local fillTypeName = thFillTypeManager:getFillTypeName(fillTypeIndex)
                            THUtils.displayMsg("Moisture price for fillType: [%s] %s", fillTypeIndex, fillTypeName)
                            THUtils.displayMsg("- Moisture percentage: %s", fillTypeMoisture)
                            THUtils.displayMsg("- Price factor: %s", priceFactor)
                            THUtils.displayMsg("- Sell price: %s", rPrice)
                            THUtils.displayMsg("- Moisture bonus: %s", moistureBonus)
                            THUtils.displayMsg("")
                        end
                    end
                end
            end)
        end
        return rPrice, ...
    end
    return appendFunc(superFunc(self, farmId, deltaFill, fillTypeIndex, toolType, extraAttributes, ...))
end
thMain:addDelayedLoadTask(function()
    THUtils.setFunctionHook("UnloadingStation", "new", THMoistureUnloadingStation, nil, false)
    THUtils.setFunctionHook("SellingStation", "update", THMoistureSellingStation)
    THUtils.setFunctionHook("SellingStation", "sellFillType", THMoistureSellingStation)
end)