-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thMain = g_thMoistureSystem
local thSettingsManager = thMain.thSettingsManager
local thFillTypeManager = thMain.thFillTypeManager
local thEffectManager = thMain.effectManager
local thAnimationManager = thMain.animationManager
local thSoundManager = thMain.soundManager
local thInputManager = thMain.thInputManager
local thI18n = thMain.i18n
THStorageDryer = {}
local THStorageDryer_mt = THUtils.createClass(THStorageDryer, Object)
local debugFlagId = THDebugUtil.createDebugFlag("THStorageDryer")
THMoistureLoadingStation.debugFlagId = debugFlagId
InitObjectClass(THStorageDryer, "THStorageDryer")
THStorageDryer.MIN_UPDATE_TIME = 500
THStorageDryer.DRYER_FX_STATE = {
    STARTING = "starting",
    RUNNING  = "running",
    STOPPING = "stopping"
}
THStorageDryer.DRYER_STATUS = {
    RUNNING = 1,
    STARTING = 2,
    STOPPING = 3,
    STOPPED = 4,
    DISABLED = 5,
    INACTIVE = 6,
    MISSING_INPUTS = 7
}
THStorageDryer.DRYER_STATUS_ID = {}
for dryerStatusId, dryerStatusIndex in pairs(THStorageDryer.DRYER_STATUS) do
    THStorageDryer.DRYER_STATUS_ID[dryerStatusIndex] = dryerStatusId
end
THStorageDryer.AUTO_STOP = {
    NEVER = 1,
    DEFAULT = 2,
    ALWAYS = 3
}
local function getCustomData(self)
    local customData = nil
    if self ~= nil then
        customData = rawget(self, self)
    end
    return customData
end
THStorageDryer.getCustomData = getCustomData
THStorageDryerActivatable = {}
local THStorageDryerActivatable_mt = THUtils.createClass(THStorageDryerActivatable)
THStorageDryerActivateEvent = {}
local THStorageDryerActivateEvent_mt = THUtils.createClass(THStorageDryerActivateEvent, Event)
InitEventClass(THStorageDryerActivateEvent, "THStorageDryerActivateEvent")
function THStorageDryer.registerXMLPaths(xmlSchema, xmlPath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.NODE_INDEX, xmlPath, "#triggerNode", "Interaction trigger node index")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#pctPerHour", "Default drying rate per hour", 0.5) -- 0.5% per hour
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#costPerHour", "Default drying costs per hour", 100)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#startDuration", "Time (in ms) for dryer to start", 1000)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, xmlPath, "#stopDuration", "Time (in ms) for dryer to stop", 1000)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.BOOL, xmlPath, "#autoStop", "Automatically stops dryer if nothing to dry", false)
    local dryerInputPath = xmlPath .. ".inputs.input(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, dryerInputPath, "#fillType", "Input fillType name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryerInputPath, "#litersPerHour", "Input liters per hour")
    local dryerOutputPath = xmlPath .. ".outputs.output(?)"
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, dryerOutputPath, "#fillType", "Output fillType name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryerOutputPath, "#pctPerHour", "Output drying rate", 1) -- 1% per hour
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryerOutputPath, "#costPerHour", "Output cost per hour", 100)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, dryerOutputPath, "#minMoisture", "Minimum drying percentage", 0)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, ".storageIds.include", "Storages used for drying (whiteList)")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, xmlPath, ".storageIds.exclude", "Storages blocked from drying (blackList)")
    for _, fxStateName in pairs(THStorageDryer.DRYER_FX_STATE) do
        local effectsPath = string.format("%s.effectNodes.%s", xmlPath, fxStateName)
        THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, effectsPath, "#effectFillType", "FillType to assign to effects")
        EffectManager.registerEffectXMLPaths(xmlSchema, effectsPath)
        local animationsPath = string.format("%s.animationNodes.%s", xmlPath, fxStateName)
        AnimationManager.registerAnimationNodesXMLPaths(xmlSchema, animationsPath)
        local samplesPath = string.format("%s.sounds", xmlPath)
        SoundManager.registerSampleXMLPaths(xmlSchema, samplesPath, fxStateName)
    end
    ObjectChangeUtil.registerObjectChangeXMLPaths(xmlSchema, xmlPath .. ".objectChanges")
end
function THStorageDryer.registerSavegameXMLPaths(xmlSchema, xmlPath)
    THUtils.registerXMLPath(xmlSchema, xmlPath, "#state", "Current storage dryer state")
end
function THStorageDryer.new(placeable, customMt)
    customMt = customMt or THStorageDryer_mt
    if THUtils.argIsValid(THUtils.getIsType(placeable, Placeable), "placeable", placeable)
        and THUtils.argIsValid(type(customMt) == "table", "customMt", customMt)
    then
        local self = Object.new(placeable.isServer, placeable.isClient, customMt)
        if self ~= nil then
            local customData = {}
            customData.rootEnvironment = thMain
            customData.parent = self
            customData.baseDirectory = placeable.baseDirectory
            customData.owningPlaceable = placeable
            customData.isOwned = false
            customData.isEnabled = true
            customData.dirtyFlag = self:getNextDirtyFlag()
            customData.dryingRate = 0.1
            customData.dryingCost = 100
            customData.cachedDryingTime = 0
            customData.totalDryingCosts = 0
            customData.autoStopDryer = false
            customData.dryerUpdateTime = 0
            customData.isDryerUpdateSafe = false
            customData.dryerStartTime = 0
            customData.dryerStartDuration = 1000
            customData.dryerStopTime = 0
            customData.dryerStopDuration = 1000
            customData.dryerStopStatus = THStorageDryer.DRYER_STATUS.STOPPED
            customData.currentDryerStatus = THStorageDryer.DRYER_STATUS.STOPPED
            customData.lastSentDryerStatus = THStorageDryer.DRYER_STATUS.STOPPED
            customData.dryerStatusText = {}
            for dryerStatusId, dryerStatusIndex in pairs(THStorageDryer.DRYER_STATUS) do
                customData.dryerStatusText[dryerStatusIndex] = thI18n:getText("thStorageDryerStatus_" .. dryerStatusId)
            end
            customData.linkedStorageIds = {
                include = {},
                exclude = {}
            }
            customData.moistureStorages = {}
            customData.moistureStoragesArray = {}
            customData.inputFillTypes = {}
            customData.outputFillTypes = {}
            customData.soundSamples = {}
            customData.animations = {}
            customData.effects = {}
            for _, fxStateName in pairs(THStorageDryer.DRYER_FX_STATE) do
                customData.animations[fxStateName] = {}
                customData.effects[fxStateName] = {}
                customData.soundSamples[fxStateName] = {}
            end
            customData.objectChanges = {}
            customData.activatable = THStorageDryerActivatable.new(self)
            if customData.activatable == nil then
                customData.isEnabled = false
                self:delete()
                return
            end
            rawset(self, self, customData)
            return self
        end
    end
end
function THStorageDryer:load(components, xmlFile, xmlKey, customEnv, i3dMappings, otherTriggerNode)
    local customData = getCustomData(self)
    local loadSuccess = false
    if customData ~= nil and customData.isEnabled then
        local placeable = customData.owningPlaceable
        local triggerNode = THUtils.getXMLValue(xmlFile, xmlKey, "#triggerNode", nil, components, i3dMappings)
        if triggerNode == nil and otherTriggerNode ~= nil then
            local triggerParent = getParent(otherTriggerNode)
            local rigidBodyType = getRigidBodyType(otherTriggerNode)
            local colMask = getCollisionMask(otherTriggerNode)
            if triggerParent ~= nil and rigidBodyType ~= nil and colMask ~= nil then
                triggerNode = clone(otherTriggerNode, false, false, false)
                link(triggerParent, triggerNode)
                setRigidBodyType(triggerNode, rigidBodyType)
                setCollisionMask(triggerNode, colMask)
            end
        end
        local pctPerHour = THUtils.getXMLValue(xmlFile, xmlKey, "#pctPerHour", 0.1)
        local costPerHour = THUtils.getXMLValue(xmlFile, xmlKey, "#costPerHour", 100)
        local startDuration = THUtils.getXMLValue(xmlFile, xmlKey, "#startDuration", 2000)
        local stopDuration = THUtils.getXMLValue(xmlFile, xmlKey, "#stopDuration", 3000)
        local autoStop = THUtils.getXMLValue(xmlFile, xmlKey, "#autoStop", false)
        if triggerNode == nil then
            THPlaceable.xmlMsg(placeable, xmlKey, "triggerNode", triggerNode, true, "Cannot find interaction trigger node")
        elseif pctPerHour <= 0 then
            THPlaceable.xmlMsg(placeable, xmlKey, "pctPerHour", pctPerHour, true, THXMLMessage.VALUE_GREATER_ZERO)
        elseif pctPerHour > 100 then
            THPlaceable.xmlMsg(placeable, xmlKey, "pctPerHour", pctPerHour, true, THXMLMessage.VALUE_LESS_EQUAL_AMOUNT, 100)
        elseif costPerHour < 0 then
            THPlaceable.xmlMsg(placeable, xmlKey, "costPerHour", costPerHour, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
        elseif startDuration < 0 then
            THPlaceable.xmlMsg(placeable, xmlKey, "startDuration", startDuration, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
        elseif stopDuration < 0 then
            THPlaceable.xmlMsg(placeable, xmlKey, "stopDuration", stopDuration, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
        else
            customData.dryingRate = pctPerHour / 100 / 60 -- convert to factor/in-game minute
            customData.dryingCost = costPerHour / 60 -- convert to cost/in-game minute
            customData.dryerStartDuration = startDuration
            customData.dryerStopDuration = stopDuration
            customData.autoStopDryer = autoStop
            local isDryerValid = true
            local inputsKey = string.format("%s.inputs", xmlKey)
            local hasInputs = false
            if xmlFile:hasProperty(inputsKey) then
                xmlFile:iterate(inputsKey .. ".input", function(_, pInputKey)
                    local fillTypeName = THUtils.getXMLValue(xmlFile, pInputKey, "#fillType")
                    local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
                    local inputLitersPerHour = THUtils.getXMLValue(xmlFile, pInputKey, "#litersPerHour", 100)
                    local isInputValid = true
                    hasInputs = true
                    if fillTypeInfo == nil or fillTypeInfo.index == FillType.UNKNOWN then
                        THPlaceable.xmlMsg(placeable, pInputKey, "fillType", fillTypeName, fillTypeName == nil, THXMLMessage.INVALID_FILL_TYPE)
                        isInputValid = false
                    elseif inputLitersPerHour < 0 then
                        THPlaceable.xmlMsg(placeable, pInputKey, "litersPerHour", inputLitersPerHour, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                        isInputValid = false
                    end
                    if isInputValid then
                        local fillTypeIndex = fillTypeInfo.index
                        local inputData = customData.inputFillTypes[fillTypeIndex]
                        if inputData ~= nil then
                            THPlaceable.xmlMsg(placeable, pInputKey, "fillType", fillTypeName, false, THXMLMessage.DUPLICATE_ENTRY)
                        else
                            inputData = {
                                fillTypeIndex = fillTypeIndex,
                                usageRate = inputLitersPerHour / 60 -- convert to liters/in-game minute
                            }
                            customData.inputFillTypes[fillTypeIndex] = inputData
                        end
                    end
                end)
            end
            if hasInputs and next(customData.inputFillTypes) == nil then
                THPlaceable.xmlMsg(placeable, inputsKey, nil, nil, true, "no valid input entries")
                isDryerValid = false
            end
            local outputsKey = string.format("%s.outputs", xmlKey)
            local hasOutputs = false
            if isDryerValid and xmlFile:hasProperty(outputsKey) then
                xmlFile:iterate(outputsKey .. ".output", function(_, pOutputKey)
                    local fillTypeName = THUtils.getXMLValue(xmlFile, pOutputKey, "#fillType")
                    local fillTypeInfo = thFillTypeManager:getFillType(fillTypeName)
                    local outputPctPerHour = THUtils.getXMLValue(xmlFile, pOutputKey, "#pctPerHour", pctPerHour)
                    local outputCostPerHour = THUtils.getXMLValue(xmlFile, pOutputKey, "#costPerHour", costPerHour)
                    local minDryPercent = THUtils.getXMLValue(xmlFile, pOutputKey, "#minMoisture", 0)
                    local isOutputValid = true
                    hasOutputs = true
                    if fillTypeInfo == nil or fillTypeInfo.index == FillType.UNKNOWN then
                        THPlaceable.xmlMsg(placeable, pOutputKey, "fillType", fillTypeName, fillTypeName == nil, THXMLMessage.INVALID_FILL_TYPE)
                        isOutputValid = false
                    elseif outputPctPerHour <= 0 then
                        THPlaceable.xmlMsg(placeable, pOutputKey, "pctPerHour", outputPctPerHour, true, THXMLMessage.VALUE_GREATER_ZERO)
                        isOutputValid = false
                    elseif outputPctPerHour > 100 then
                        THPlaceable.xmlMsg(placeable, pOutputKey, "pctPerHour", outputPctPerHour, true, THXMLMessage.VALUE_LESS_EQUAL_AMOUNT, 100)
                        isOutputValid = false
                    elseif outputCostPerHour < 0 then
                        THPlaceable.xmlMsg(placeable, pOutputKey, "costPerHour", outputCostPerHour, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
                        isOutputValid = false
                    elseif minDryPercent < 0 or minDryPercent > 100 then
                        THPlaceable.xmlMsg(placeable, pOutputKey, "minPercent", minDryPercent, true, THXMLMessage.VALUE_OUT_OF_RANGE, 0, 100)
                        isOutputValid = false
                    end
                    if isOutputValid then
                        local fillTypeIndex = fillTypeInfo.index
                        local outputData = customData.outputFillTypes[fillTypeIndex]
                        if outputData ~= nil then
                            THPlaceable.xmlMsg(placeable, pOutputKey, "fillType", fillTypeName, false, THXMLMessage.DUPLICATE_ENTRY)
                        else
                            outputData = {
                                fillTypeIndex = fillTypeIndex,
                                dryingRate = outputPctPerHour / 100 / 60, -- convert to factor/in-game minute
                                dryingCost = outputCostPerHour / 60, -- convert to cost/in-game minute
                                minPercent = minDryPercent / 100 -- convert to factor
                            }
                            customData.outputFillTypes[fillTypeIndex] = outputData
                        end
                    end
                end)
            end
            if hasOutputs and next(customData.outputFillTypes) == nil then
                THPlaceable.xmlMsg(placeable, outputsKey, nil, nil, true, "no valid outputs found")
                isDryerValid = false
            end
            if isDryerValid then
                local storageIdsKey = string.format("%s.storageIds", xmlKey)
                local includeStorageIds = THUtils.getXMLValue(xmlFile, storageIdsKey, ".include", "")
                local excludeStorageIds = THUtils.getXMLValue(xmlFile, storageIdsKey, ".exclude", "")
                local function loadLinkedStorageIds(pIdString, pDataTable)
                    return THUtils.call(function()
                        if pIdString ~= "" then
                            local storageIdsArray = THUtils.splitString(pIdString, " ")
                            if storageIdsArray ~= nil and #storageIdsArray > 0 then
                                for _, storageId in pairs(storageIdsArray) do
                                    if storageId ~= "" then
                                        pDataTable[storageId:upper()] = true
                                    end
                                end
                            end
                        end
                        return true
                    end)
                end
                if not loadLinkedStorageIds(includeStorageIds, customData.linkedStorageIds.include)
                    or not loadLinkedStorageIds(excludeStorageIds, customData.linkedStorageIds.exclude)
                then
                    isDryerValid = false
                end
            end
            if isDryerValid and self.isClient then
                local samplesKey = string.format("%s.sounds", xmlKey)
                if not xmlFile:hasProperty(samplesKey) then
                    for _, fxStateName in pairs(THStorageDryer.DRYER_FX_STATE) do
                        local sampleTypeKey = string.format("%s.%s", samplesKey, fxStateName)
                        local sampleTemplate = nil
                        local sampleLinkNode = "0>"
                        if fxStateName == THStorageDryer.DRYER_FX_STATE.RUNNING then
                            sampleTemplate = "STRAWBLOWER_WORK"
                        elseif fxStateName == THStorageDryer.DRYER_FX_STATE.STARTING then
                            sampleTemplate = "STRAWBLOWER_START"
                        elseif fxStateName == THStorageDryer.DRYER_FX_STATE.STOPPING then
                            sampleTemplate = "STRAWBLOWER_STOP"
                        end
                        if sampleTemplate ~= nil then
                            xmlFile:setString(sampleTypeKey .. "#template", sampleTemplate)
                            xmlFile:setString(sampleTypeKey .. "#linkNode", sampleLinkNode)
                        end
                    end
                end
                for _, fxStateName in pairs(THStorageDryer.DRYER_FX_STATE) do
                    local effectsKey = string.format("%s.effectNodes.%s", xmlKey, fxStateName)
                    local effectsTable = thEffectManager:loadEffect(xmlFile, effectsKey, components, placeable, i3dMappings)
                    if effectsTable ~= nil then
                        local effectFillType = THUtils.getXMLValue(xmlFile, effectsKey, "#effectFillType")
                        local effectFillTypeInfo = nil
                        if effectFillType ~= nil then
                            effectFillTypeInfo = thFillTypeManager:getFillType(effectFillType)
                            if effectFillTypeInfo == nil then
                                THPlaceable.xmlMsg(placeable, effectsKey, "effectFillType", effectFillType, false, THXMLMessage.INVALID_FILL_TYPE)
                            end
                        end
                        if effectFillTypeInfo ~= nil then
                            thEffectManager:setFillType(effectsTable, effectFillTypeInfo.index)
                        else
                            thEffectManager:setFillType(effectsTable, FillType.UNKNOWN)
                        end
                        customData.effects[fxStateName] = effectsTable
                    end
                    local animationsKey = string.format("%s.animationNodes.%s", xmlKey, fxStateName)
                    customData.animations[fxStateName] = thAnimationManager:loadAnimations(xmlFile, animationsKey, components, placeable, i3dMappings)
                    local loopCount = 1
                    if fxStateName == THStorageDryer.DRYER_FX_STATE.RUNNING then
                        loopCount = 0
                    end
                    thSoundManager:loadSamplesFromXML(xmlFile, samplesKey, fxStateName, customData.baseDirectory, components, loopCount, AudioGroup.ENVIRONMENT, i3dMappings, nil, customData.soundSamples[fxStateName])
                end
            end
            if isDryerValid then
                customData.interactionTrigger = triggerNode
                addTrigger(triggerNode, "storageDryerTriggerCallback", self)
                loadSuccess = true
            end
        end
        if loadSuccess then
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THPlaceable.displayMsg(placeable, "Storage dryer loaded successfully")
                THPlaceable.displayMsg(placeable, "- XML Key: %s", xmlKey)
                THUtils.printTable(customData, 1)
                THPlaceable.displayMsg(placeable, "- Animations:")
                THUtils.printTable(customData.animations, 1)
                THPlaceable.displayMsg(placeable, "- Effects:")
                THUtils.printTable(customData.effects, 1)
                THPlaceable.displayMsg(placeable, "- Sounds:")
                THUtils.printTable(customData.soundSamples, 1)
            end
        end
    end
    return loadSuccess
end
function THStorageDryer:update(dt, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil and customData.isEnabled then
            THUtils.call(function()
                if self.isServer then
                    if customData.dryerUpdateTime >= THStorageDryer.MIN_UPDATE_TIME then
                        customData.isDryerUpdateSafe = true
                    else
                        customData.dryerUpdateTime = customData.dryerUpdateTime + dt
                        customData.isDryerUpdateSafe = false
                        self:raiseActive()
                    end
                end
                if customData.currentDryerStatus == THStorageDryer.DRYER_STATUS.STARTING then
                    customData.dryerStopTime = 0
                    if customData.dryerStartTime > 0 then
                        customData.dryerStartTime = customData.dryerStartTime - dt
                        self:raiseActive()
                    else
                        customData.dryerStartTime = 0
                        self:setDryerStatus(THStorageDryer.DRYER_STATUS.RUNNING, true)
                    end
                elseif customData.currentDryerStatus == THStorageDryer.DRYER_STATUS.STOPPING then
                    customData.dryerStartTime = 0
                    if customData.dryerStopTime > 0 then
                        customData.dryerStopTime = customData.dryerStopTime - dt
                        self:raiseActive()
                    else
                        customData.dryerStopTime = 0
                        self:setDryerStatus(customData.dryerStopStatus, true)
                    end
                else
                    customData.dryerStartTime = 0
                    customData.dryerStopTime = 0
                end
            end)
        end
        return ...
    end
    return appendFunc(THStorageDryer:superClass().update(self, dt, ...))
end
function THStorageDryer:readStream(streamId, connection, ...)
    local function appendFunc(...)
        if connection:getIsServer() then
            self:readDryerDataFromStream(streamId, connection)
        end
        return ...
    end
    return appendFunc(THStorageDryer:superClass().readStream(self, streamId, connection, ...))
end
function THStorageDryer:writeStream(streamId, connection, ...)
    local function appendFunc(...)
        if not connection:getIsServer() then
            self:writeDryerDataToStream(streamId, connection)
        end
        return ...
    end
    return appendFunc(THStorageDryer:superClass().writeStream(self, streamId, connection, ...))
end
function THStorageDryer:readUpdateStream(streamId, timestamp, connection, ...)
    local function appendFunc(...)
        if connection:getIsServer() then
            if streamReadBool(streamId) then -- customData exists
                if streamReadBool(streamId) then -- dryer isDirty
                    self:readDryerDataFromStream(streamId, connection)
                end
            end
        end
        return ...
    end
    return appendFunc(THStorageDryer:superClass().readUpdateStream(self, streamId, timestamp, connection, ...))
end
function THStorageDryer:writeUpdateStream(streamId, connection, dirtyMask, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if not connection:getIsServer() then
            if streamWriteBool(streamId, customData ~= nil) then
                if streamWriteBool(streamId, bitAND(dirtyMask, customData.dirtyFlag) ~= 0) then
                    self:writeDryerDataToStream(streamId, connection)
                end
            end
        end
        return ...
    end
    return appendFunc(THStorageDryer:superClass().writeUpdateStream(self, streamId, connection, dirtyMask, ...))
end
function THStorageDryer:delete()
    local customData = getCustomData(self)
    if customData ~= nil then
        customData.isEnabled = false
    end
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                self:setDryerStatus(THStorageDryer.DRYER_STATUS.DISABLED, true)
                if customData.interactionTrigger ~= nil then
                    removeTrigger(customData.interactionTrigger)
                    customData.interactionTrigger = nil
                end
                for _, fxStateName in pairs(THStorageDryer.DRYER_FX_STATE) do
                    if customData.effects[fxStateName] ~= nil then
                        thEffectManager:deleteEffects(customData.effects[fxStateName])
                    end
                    if customData.animations[fxStateName] ~= nil then
                        thAnimationManager:deleteAnimations(customData.animations[fxStateName])
                    end
                    if customData.soundSamples[fxStateName] ~= nil then
                        thSoundManager:deleteSamples(customData.soundSamples[fxStateName])
                    end
                end
                local mission = thMain.mission
                if mission ~= nil then
                    if customData.activatable ~= nil then
                        mission.activatableObjectsSystem:removeActivatable(customData.activatable)
                    end
                end
            end)
        end
        return ...
    end
    return appendFunc(THStorageDryer:superClass().delete(self))
end
function THStorageDryer:readDryerDataFromStream(streamId, connection)
    local customData = getCustomData(self)
    if connection:getIsServer() then
        local dryerStatus = THStorageDryer.DRYER_STATUS.STOPPED
        if streamReadBool(streamId) then -- server customData exists/enabled
            dryerStatus = streamReadUIntN(streamId, 4)
        elseif customData ~= nil then
            customData.isEnabled = false
        end
        self:setDryerStatus(dryerStatus, true)
    end
end
function THStorageDryer:writeDryerDataToStream(streamId, connection)
    local customData = getCustomData(self)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, customData ~= nil and customData.isEnabled == true) then
            local sentDryerStatus = customData.currentDryerStatus
            if sentDryerStatus == THStorageDryer.DRYER_STATUS.STOPPING
                and customData.dryerStopStatus ~= nil
            then
                sentDryerStatus = customData.dryerStopStatus
            end
            streamWriteUIntN(streamId, sentDryerStatus, 4)
            customData.lastSentDryerStatus = sentDryerStatus
            THUtils.call(function()
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local placeable = customData.owningPlaceable
                    local dryerStatusId = THStorageDryer.DRYER_STATUS_ID[sentDryerStatus]
                    THPlaceable.displayMsg(placeable, "Sent dryer status: %s", dryerStatusId)
                end
            end)
        end
    end
end
function THStorageDryer:storageDryerTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId, ...)
    local customData = getCustomData(self)
    if customData ~= nil then
        THUtils.call(function()
            local mission = thMain.mission
            local player = mission.player
            local activatableObjectsSystem = mission.activatableObjectsSystem
            if (onEnter or onLeave) and player ~= nil and player.rootNode == otherId then
                customData.activatable:updateActionTexts()
                if onEnter and customData.isEnabled then
                    activatableObjectsSystem:addActivatable(customData.activatable)
                end
                if onLeave then
                    activatableObjectsSystem:removeActivatable(customData.activatable)
                end
            end
        end)
    end
end
function THStorageDryer:setOwnerFarmId(farmId, noEventSend, ...)
    local customData = getCustomData(self)
    if customData ~= nil and customData.isEnabled then
        THUtils.call(function()
            self:processDryingCosts()
        end)
    end
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                customData.isOwned = farmId ~= AccessHandler.EVERYONE
            end)
            return ...
        end
    end
    return appendFunc(THStorageDryer:superClass().setOwnerFarmId(self, farmId, noEventSend, ...))
end
function THStorageDryer:getIsEnabled()
    local customData = getCustomData(self)
    return customData ~= nil and customData.isEnabled == true
end
function THStorageDryer:getDryerStatus()
    local customData = getCustomData(self)
    if customData ~= nil and customData.isEnabled
        and customData.currentDryerStatus ~= nil
        and customData.currentDryerStatus > 0
    then
        return customData.currentDryerStatus
    end
    return THStorageDryer.DRYER_STATUS.DISABLED
end
function THStorageDryer:getDryerStatusText()
    local customData = getCustomData(self)
    local statusTextUnknown = "Unknown"
    if customData ~= nil then
        return customData.dryerStatusText[customData.currentDryerStatus] or statusTextUnknown
    end
    return statusTextUnknown
end
function THStorageDryer:getDryerStorages(farmId, refresh)
    local customData = getCustomData(self)
    if customData ~= nil and customData.isEnabled then
        if refresh == true then
            THUtils.clearTable(customData.moistureStorages)
            THUtils.clearTable(customData.moistureStoragesArray)
            local placeable = customData.owningPlaceable
            if placeable ~= nil and placeable.thAA_getPlaceableObjects ~= nil then
                local storagesArray, numStorages = placeable:thAA_getPlaceableObjects(THPlaceable.OBJECT_TYPE.STORAGE, farmId)
                if storagesArray ~= nil and numStorages > 0 then
                    local storage, allowAddStorage = nil, nil
                    for storageIndex = 1, numStorages do
                        storage = storagesArray[storageIndex]
                        allowAddStorage = false
                        if customData.moistureStorages[storage] == nil then
                            if next(customData.linkedStorageIds.include) ~= nil
                                or next(customData.linkedStorageIds.exclude) ~= nil
                            then
                                local thStorage = THPlaceable.getObjectData(storage)
                                if next(customData.linkedStorageIds.include) ~= nil then
                                    if thStorage ~= nil and thStorage.id ~= nil
                                        and customData.linkedStorageIds.include[thStorage.id]
                                    then
                                        allowAddStorage = true
                                    end
                                else
                                    if thStorage == nil or thStorage.id == nil
                                        or not customData.linkedStorageIds.exclude[thStorage.id]
                                    then
                                        allowAddStorage = true
                                    end
                                end
                            else
                                allowAddStorage = true
                            end
                        end
                        if allowAddStorage then
                            table.insert(customData.moistureStoragesArray, storage)
                            customData.moistureStorages[storage] = storage
                        end
                    end
                end
            end
        end
        return customData.moistureStoragesArray, #customData.moistureStoragesArray
    end
    return nil, 0
end
function THStorageDryer:setDryerStatus(statusIndex, noEventSend)
    local customData = getCustomData(self)
    if customData ~= nil and (self.isServer or noEventSend == true) then
        local placeable = customData.owningPlaceable
        local currentStatus = customData.currentDryerStatus
        local statusId = THStorageDryer.DRYER_STATUS_ID[statusIndex]
        if not customData.isEnabled or statusId == nil then
            if statusId == nil then
                THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "statusIndex", statusIndex)
            end
            statusIndex = THStorageDryer.DRYER_STATUS.DISABLED
        end
        if statusIndex ~= currentStatus or noEventSend == true then
            local currentIsRunning = currentStatus == THStorageDryer.DRYER_STATUS.RUNNING
                or currentStatus == THStorageDryer.DRYER_STATUS.STARTING
            local nextIsRunning = statusIndex == THStorageDryer.DRYER_STATUS.RUNNING
                or statusIndex == THStorageDryer.DRYER_STATUS.STARTING
            if nextIsRunning ~= currentIsRunning then
                if nextIsRunning then -- dryer starting up
                    customData.dryerStartTime = customData.dryerStartDuration
                    customData.dryerStopTime = 0
                    statusIndex = THStorageDryer.DRYER_STATUS.STARTING
                else -- dryer shutting down
                    customData.dryerStartTime = 0
                    customData.dryerStopTime = customData.dryerStopDuration
                    customData.dryerStopStatus = statusIndex
                    if statusIndex == THStorageDryer.DRYER_STATUS.STOPPING then
                        customData.dryerStopStatus = THStorageDryer.DRYER_STATUS.STOPPED
                    end
                    statusIndex = THStorageDryer.DRYER_STATUS.STOPPING
                end
            end
            if statusIndex ~= currentStatus or noEventSend == true then
                if THDebugUtil.getIsEnabled(debugFlagId) then
                    local currentStatusId = THStorageDryer.DRYER_STATUS_ID[currentStatus]
                    local nextStatusId = THStorageDryer.DRYER_STATUS_ID[statusIndex]
                    THPlaceable.displayMsg(placeable, "Dryer status changed from %q to %q", currentStatusId, nextStatusId)
                end
                customData.currentDryerStatus = statusIndex
                customData.activatable:updateActionTexts()
                self:updateDryerFX()
                self:raiseActive()
            end
            if self.isServer and noEventSend ~= true then
                if customData.currentDryerStatus ~= customData.lastSentDryerStatus then
                    self:raiseDirtyFlags(customData.dirtyFlag)
                end
            end
        end
    end
end
function THStorageDryer:updateDryer(elapsedMinutes)
    elapsedMinutes = elapsedMinutes or 0
    local customData = getCustomData(self)
    if self.isServer and customData ~= nil and customData.isEnabled then
        if not customData.isDryerUpdateSafe then
            customData.cachedDryingTime = customData.cachedDryingTime + elapsedMinutes
        else
            local storageList = self:getDryerStorages(nil, true)
            local currentDryerStatus = customData.currentDryerStatus
            local globalDryingFactor = thSettingsManager:getSettingValue(THMoistureSystem.SETTING.STORAGE_DRYING_FACTOR) or 1
            elapsedMinutes = (elapsedMinutes + customData.cachedDryingTime) * globalDryingFactor
            customData.dryerUpdateTime = 0
            customData.cachedDryingTime = 0
            customData.isDryerUpdateSafe = false
            if storageList == nil or next(storageList) == nil then
                currentDryerStatus = THStorageDryer.DRYER_STATUS.DISABLED
            elseif currentDryerStatus == THStorageDryer.DRYER_STATUS.RUNNING then
                local baseDryingRate = customData.dryingRate * elapsedMinutes
                local baseDryingCost = customData.dryingCost * elapsedMinutes
                local totalDryingFactor = 1
                local totalDryingCost = 0
                for inputFillType, inputData in pairs(customData.inputFillTypes) do
                    local totalUsage = inputData.usageRate * elapsedMinutes
                    local remainingUsage = totalUsage
                    if totalUsage > 0 then
                        for _, storage in pairs(storageList) do
                            local oldFillLevel = storage:getFillLevel(inputFillType)
                            if oldFillLevel > 0 then
                                local newFillLevel = math.max(0, oldFillLevel - remainingUsage)
                                storage:setFillLevel(newFillLevel, inputFillType)
                                if oldFillLevel >= remainingUsage then
                                    remainingUsage = 0
                                else
                                    newFillLevel = storage:getFillLevel(inputFillType)
                                    remainingUsage = math.max(0, remainingUsage - (oldFillLevel - newFillLevel))
                                end
                                if remainingUsage <= 0 then
                                    break
                                end
                            end
                        end
                        local appliedUsage = totalUsage - remainingUsage
                        local dryingFactor = MathUtil.clamp(appliedUsage / totalUsage, 0, 1)
                        totalDryingFactor = math.min(dryingFactor, totalDryingFactor)
                    end
                end
                if totalDryingFactor <= 0 then
                    currentDryerStatus = THStorageDryer.DRYER_STATUS.MISSING_INPUTS
                else
                    local numOutputs = 0
                    local isMoistureChanged = false
                    for _, storage in pairs(storageList) do
                        local thMoistureStorage = THMoistureStorage.getCustomData(storage)
                        local storageFillLevels = storage:getFillLevels()
                        if thMoistureStorage ~= nil and storageFillLevels ~= nil then
                            for fillTypeIndex, fillLevel in pairs(storageFillLevels) do
                                local fillTypeMoistureData = thMain:getFillTypeMoistureData(fillTypeIndex)
                                if fillTypeMoistureData ~= nil then
                                    local storageMoisture = thMoistureStorage:getFillTypeMoisture(fillTypeIndex)
                                    if fillLevel > 0
                                        and storageMoisture ~= nil and storageMoisture >= 0
                                    then
                                        local outputData = customData.outputFillTypes[fillTypeIndex]
                                        local outputDryingRate = baseDryingRate
                                        local outputDryingCost = baseDryingCost
                                        local minMoisture = fillTypeMoistureData.minPercent
                                        if outputData ~= nil then
                                            outputDryingRate = outputData.dryingRate * elapsedMinutes
                                            outputDryingCost = outputData.dryingCost * elapsedMinutes
                                            minMoisture = outputData.minPercent
                                        end
                                        outputDryingRate = outputDryingRate * totalDryingFactor
                                        outputDryingCost = outputDryingCost * totalDryingFactor
                                        if storageMoisture > minMoisture then
                                            local newMoisture = math.max(minMoisture, storageMoisture - outputDryingRate)
                                            if newMoisture ~= storageMoisture then
                                                thMoistureStorage:setFillTypeMoisture(fillTypeIndex, newMoisture)
                                                isMoistureChanged = true
                                            end
                                        end
                                        totalDryingCost = totalDryingCost + outputDryingCost
                                    end
                                    numOutputs = numOutputs + 1
                                end
                            end
                        end
                    end
                    if numOutputs == 0 then
                        currentDryerStatus = THStorageDryer.DRYER_STATUS.DISABLED
                    else
                        totalDryingCost = totalDryingCost / numOutputs
                        if not isMoistureChanged then
                            local globalAutoStop = thSettingsManager:getSettingValue(THMoistureSystem.SETTING.AUTO_STOP_DRYERS) or THStorageDryer.AUTO_STOP.DEFAULT
                            local autoStopDryer = customData.autoStopDryer
                            if globalAutoStop == THStorageDryer.AUTO_STOP.ALWAYS then
                                autoStopDryer = true
                            elseif globalAutoStop == THStorageDryer.AUTO_STOP.NEVER then
                                autoStopDryer = false
                            end
                            if autoStopDryer then
                                currentDryerStatus = THStorageDryer.DRYER_STATUS.INACTIVE
                            end
                        end
                    end
                end
                if totalDryingCost > 0 then
                    customData.totalDryingCosts = customData.totalDryingCosts + totalDryingCost
                end
            end
            self:setDryerStatus(currentDryerStatus)
        end
        self:raiseActive()
    end
end
function THStorageDryer:updateDryerFX()
    local customData = getCustomData(self)
    if customData ~= nil then
        if self.isClient then
            local dryerStatus = customData.currentDryerStatus
            if not customData.isEnabled then
                dryerStatus = THStorageDryer.DRYER_STATUS.DISABLED
            end
            if dryerStatus == THStorageDryer.DRYER_STATUS.STARTING
                or dryerStatus == THStorageDryer.DRYER_STATUS.RUNNING
            then
                thEffectManager:stopEffects(customData.effects.stopping)
                thAnimationManager:stopAnimations(customData.animations.stopping)
                thSoundManager:stopSamples(customData.soundSamples.stopping)
                if dryerStatus == THStorageDryer.DRYER_STATUS.STARTING then
                    thEffectManager:stopEffects(customData.effects.running)
                    thAnimationManager:stopAnimations(customData.animations.running)
                    thSoundManager:stopSamples(customData.soundSamples.running)
                    thEffectManager:startEffects(customData.effects.starting)
                    thAnimationManager:startAnimations(customData.animations.starting)
                    thSoundManager:playSamples(customData.soundSamples.starting)
                else
                    thEffectManager:stopEffects(customData.effects.starting)
                    thAnimationManager:stopAnimations(customData.animations.starting)
                    thSoundManager:stopSamples(customData.soundSamples.starting)
                    thEffectManager:startEffects(customData.effects.running)
                    thAnimationManager:startAnimations(customData.animations.running)
                    thSoundManager:playSamples(customData.soundSamples.running)
                end
            else
                thEffectManager:stopEffects(customData.effects.starting)
                thEffectManager:stopEffects(customData.effects.running)
                thAnimationManager:stopAnimations(customData.animations.starting)
                thAnimationManager:stopAnimations(customData.animations.running)
                thSoundManager:stopSamples(customData.soundSamples.starting)
                thSoundManager:stopSamples(customData.soundSamples.running)
                if customData.currentDryerStatus == THStorageDryer.DRYER_STATUS.STOPPING then
                    thEffectManager:startEffects(customData.effects.stopping)
                    thAnimationManager:startAnimations(customData.animations.stopping)
                    thSoundManager:playSamples(customData.soundSamples.stopping)
                else
                    thEffectManager:stopEffects(customData.effects.stopping)
                    thAnimationManager:stopAnimations(customData.animations.stopping)
                    thSoundManager:stopSamples(customData.soundSamples.stopping)
                end
            end
        end
    end
end
function THStorageDryer:processDryingCosts()
    local customData = getCustomData(self)
    if customData ~= nil and customData.isEnabled
        and self.isServer and customData.isOwned
    then
        local mission = thMain.mission
        local totalDryingCosts = customData.totalDryingCosts or 0
        if mission ~= nil and totalDryingCosts > 0 then
            local ownerFarmId = self:getOwnerFarmId()
            local moneyChangeType = thMain.moneyChangeTypes.dryingCosts
            mission:addMoney(-totalDryingCosts, ownerFarmId, moneyChangeType, true, true)
            customData.totalDryingCosts = 0
        end
    end
end
function THStorageDryerActivatable.new(dryer, customMt)
    customMt = customMt or THStorageDryerActivatable_mt
    if THUtils.argIsValid(THUtils.getIsType(dryer, THStorageDryer), "dryer", dryer)
        and THUtils.argIsValid(type(customMt) == "table", "customMt", customMt)
    then
        local self = setmetatable({}, customMt)
        self.isClient = dryer.isClient
        self.isServer = dryer.isServer
        self.storageDryer = dryer
        self.activateText = ""
        return self
    end
end
function THStorageDryerActivatable:getIsActivatable()
    local dryerData = getCustomData(self.storageDryer)
    if dryerData ~= nil and dryerData.isEnabled then
        if dryerData.currentDryerStatus ~= THStorageDryer.DRYER_STATUS.DISABLED then
            if not dryerData.isOwned then
                return true
            else
                local mission = thMain.mission
                if mission ~= nil then
                    local missionFarmId = mission:getFarmId()
                    if mission.accessHandler:canFarmAccess(missionFarmId, self.storageDryer) then
                        return true
                    end
                end
            end
        end
    end
    return false
end
function THStorageDryerActivatable:registerCustomInput(inputContext)
    local dryerData = getCustomData(self.storageDryer)
    if dryerData ~= nil and dryerData.isEnabled then
        local _, actionEventId = thInputManager:registerActionEvent("THAA_PLAYER_FUNCTION", 1, self, self.onToggleDryer, false, true, false, true)
        if actionEventId ~= nil then
            self.actionIdToggleDryer = actionEventId
            thInputManager:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_HIGH)
            thInputManager:setActionEventTextVisibility(actionEventId, true)
        end
        self:updateActionTexts()
    end
end
function THStorageDryerActivatable:removeCustomInput(inputContext)
    thInputManager:removeActionEventsByTarget(self)
    self.actionIdToggleDryer = nil
end
function THStorageDryerActivatable:updateActionTexts()
    if self.actionIdToggleDryer ~= nil then
        local dryerData = getCustomData(self.storageDryer)
        local isDryerActive = false
        local activateText = ""
        if self:getIsActivatable() and dryerData ~= nil then
            if dryerData.currentDryerStatus == THStorageDryer.DRYER_STATUS.RUNNING then
                activateText = thI18n:getText("thAction_deactivateStorageDryer")
            else
                activateText = thI18n:getText("thAction_activateStorageDryer")
            end
            isDryerActive = true
        end
        thInputManager:setActionEventText(self.actionIdToggleDryer, activateText)
        thInputManager:setActionEventTextVisibility(self.actionIdToggleDryer, isDryerActive)
        thInputManager:setActionEventActive(self.actionIdToggleDryer, isDryerActive)
    end
end
function THStorageDryerActivatable:onToggleDryer()
    local dryerData = getCustomData(self.storageDryer)
    if dryerData ~= nil and dryerData.isEnabled then
        local nextIsActive = true
        if dryerData.currentDryerStatus == THStorageDryer.DRYER_STATUS.RUNNING
            or dryerData.currentDryerStatus == THStorageDryer.DRYER_STATUS.STARTING
        then
            nextIsActive = false
        end
        if self.isServer then
            if nextIsActive == true then
                self.storageDryer:setDryerStatus(THStorageDryer.DRYER_STATUS.RUNNING)
            else
                self.storageDryer:setDryerStatus(THStorageDryer.DRYER_STATUS.STOPPED)
            end
        else
            THStorageDryerActivateEvent.sendEvent(self.storageDryer, nextIsActive)
        end
    end
end
function THStorageDryerActivateEvent.emptyNew()
    return Event.new(THStorageDryerActivateEvent_mt)
end
function THStorageDryerActivateEvent.new(storageDryer, isActive)
    local self = THStorageDryerActivateEvent.emptyNew()
    self.isEventValid = false
    if THUtils.argIsValid(THUtils.getIsType(storageDryer, THStorageDryer), "storageDryer", storageDryer)
        and THUtils.argIsValid(type(isActive) == "boolean", "isActive", isActive)
    then
        self.storageDryer = storageDryer
        self.isActive = isActive
        self.isEventValid = true
    end
    return self
end
function THStorageDryerActivateEvent:readStream(streamId, connection)
    if not connection:getIsServer() then -- server reading from client
        self.isEventValid = streamReadBool(streamId)
        if self.isEventValid then
            self.storageDryer = NetworkUtil.readNodeObject(streamId)
            self.isActive = streamReadBool(streamId)
            self:run(connection)
        end
    else
        THUtils.errorMsg(false, "THStorageDryerActivateEvent should not read from clients!")
    end
end
function THStorageDryerActivateEvent:writeStream(streamId, connection)
    if connection:getIsServer() then -- client sending to server
        if streamWriteBool(streamId, self.isEventValid) then
            NetworkUtil.writeNodeObject(streamId, self.storageDryer)
            streamWriteBool(streamId, self.isActive)
        end
    else
        THUtils.errorMsg(false, "THStorageDryerActivateEvent should not write from server!")
    end
end
function THStorageDryerActivateEvent:run(connection)
    if not connection:getIsServer() then -- server reading from client
        if self.isEventValid then
            if self.isActive == true then
                self.storageDryer:setDryerStatus(THStorageDryer.DRYER_STATUS.RUNNING)
            else
                self.storageDryer:setDryerStatus(THStorageDryer.DRYER_STATUS.STOPPED)
            end
        end
    end
end
function THStorageDryerActivateEvent.sendEvent(storageDryer, isActive)
    THUtils.call(function()
        if g_server == nil and g_client ~= nil then
            local newEvent = THStorageDryerActivateEvent.new(storageDryer, isActive)
            if newEvent ~= nil and newEvent.isEventValid then
                g_client:getServerConnection():sendEvent(newEvent)
            end
        else
            THUtils.stackMsg("THStorageDryerActivateEvent should not be sent from the server!")
        end
    end)
end