-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thMain = g_thMoistureSystem
local thModName = thMain.modName
local thSettingsManager = thMain.thSettingsManager
local thFillTypeManager = thMain.thFillTypeManager
local thI18n = thMain.i18n
THPlaceableMoistureStorage = {}
local debugFlagId = THDebugUtil.createDebugFlag("THPlaceableMoistureStorage")
THPlaceableMoistureStorage.debugFlagId = debugFlagId
THPlaceableMoistureStorage.MOD_NAME = thModName
THPlaceableMoistureStorage.SPEC_NAME = "moistureStorage"
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THPlaceableMoistureStorage.SPEC_NAME, thModName, verbose)
end
THPlaceableMoistureStorage.getSpecTable = getSpecTable
function THPlaceableMoistureStorage.initSpecialization()
end
function THPlaceableMoistureStorage.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THPlaceable, specializations)
end
function THPlaceableMoistureStorage.registerXMLPaths(xmlSchema, xmlPath)
    xmlSchema:setXMLSpecializationType("THMoistureStorage")
    local basePath = THPlaceable.getConfigSchemaPath(THPlaceableMoistureStorage)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, basePath, "#decayFactor", 1)
    THStorageDryer.registerXMLPaths(xmlSchema, basePath .. ".storageDryer")
    xmlSchema:setXMLSpecializationType()
end
function THPlaceableMoistureStorage.registerSavegameXMLPaths(xmlSchema, xmlPath)
    xmlSchema:setXMLSpecializationType("THMoistureStorage")
    local basePath = THPlaceable.getSavegameSchemaPath(THPlaceableMoistureStorage)
    local fillTypeLossPath = string.format("%s.storageLoss.fillType(?)", basePath)
    THUtils.registerXMLPath(xmlSchema, XMLValueType.STRING, fillTypeLossPath, "#name", "FillType name")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeLossPath, "#amount", "Total loss for this fillType")
    THUtils.registerXMLPath(xmlSchema, XMLValueType.FLOAT, fillTypeLossPath, "#factor", "Total loss factor for this fillType")
    THStorageDryer.registerSavegameXMLPaths(xmlSchema, basePath .. ".storageDryer")
    xmlSchema:setXMLSpecializationType()
end
function THPlaceableMoistureStorage.registerEvents(placeableType)
end
function THPlaceableMoistureStorage.registerEventListeners(placeableType)
    SpecializationUtil.registerEventListener(placeableType, "onPreLoad", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onLoad", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onFinalizePlacement", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onDelete", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onReadStream", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onWriteStream", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onReadUpdateStream", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onWriteUpdateStream", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onDraw", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onMinuteChanged", THPlaceableMoistureStorage)
    SpecializationUtil.registerEventListener(placeableType, "onDayChanged", THPlaceableMoistureStorage)
end
function THPlaceableMoistureStorage.registerFunctions(placeableType)
    SpecializationUtil.registerFunction(placeableType, "thAA_readMoistureStorageFromStream", THPlaceableMoistureStorage.thAA_readMoistureStorageFromStream)
    SpecializationUtil.registerFunction(placeableType, "thAA_writeMoistureStorageToStream", THPlaceableMoistureStorage.thAA_writeMoistureStorageToStream)
    SpecializationUtil.registerFunction(placeableType, "thAA_clearMoistureStorageLoss", THPlaceableMoistureStorage.thAA_clearMoistureStorageLoss)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateMoistureStorageLoss", THPlaceableMoistureStorage.thAA_updateMoistureStorageLoss)
    SpecializationUtil.registerFunction(placeableType, "thAA_updateMoistureStorageBuffer", THPlaceableMoistureStorage.thAA_updateMoistureStorageBuffer)
    SpecializationUtil.registerFunction(placeableType, "thAA_showMoistureHudBoxEntries", THPlaceableMoistureStorage.thAA_showMoistureHudBoxEntries)
end
function THPlaceableMoistureStorage.registerOverwrittenFunctions(placeableType)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "setOwnerFarmId", THPlaceableMoistureStorage.setOwnerFarmId)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "collectPickObjects", THPlaceableMoistureStorage.collectPickObjects)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "getNeedMinuteChanged", THPlaceableMoistureStorage.getNeedMinuteChanged)
    SpecializationUtil.registerOverwrittenFunction(placeableType, "getNeedDayChanged", THPlaceableMoistureStorage.getNeedDayChanged)
end
function THPlaceableMoistureStorage:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.moistureBuffer = {}
    specTable.moistureBufferMapping = {}
    if specTable.moistureHudBox == nil then
        specTable.moistureHudBox = thMain.hud.infoDisplay:createBox(KeyValueInfoHUDBox)
    end
    specTable.fillTypeLoss = {}
    specTable.lossFillTypes = {}
    specTable.storageDecayFactor = 1
    specTable.isMoistureHudBoxEnabled = true
    specTable.hasDisplayMoisture = false
    specTable.hasDisplayLoss = false
    specTable.dryerRunTime = 0
    specTable.dirtyFlag = self:getNextDirtyFlag()
    specTable.hudTitles = {
        storage = thI18n:getText("thUI_storage"),
        moisture = thI18n:getText("thUI_moisture"),
        storageLoss = thI18n:getText("thUI_storageLoss"),
        dryerStatus = thI18n:getText("thStorageDryerStatus")
    }
end
function THPlaceableMoistureStorage:onLoad(savegame)
    local specTable = getSpecTable(self)
    local siloSpec = self.spec_silo
    local infoTriggerSpec = self.spec_infoTrigger
    local components = self.components
    local xmlFile = self.xmlFile
    local xmlBaseKey = THPlaceable.getConfigBaseKey(THPlaceableMoistureStorage)
    local customEnv = self.customEnvironment
    local i3dMappings = self.i3dMappings
    specTable.storageDecayFactor = THUtils.getXMLValue(xmlFile, xmlBaseKey, "#decayFactor", 1)
    if specTable.storageDecayFactor < 0 then
        THPlaceable.xmlMsg(self, xmlBaseKey, "decayFactor", specTable.storageDecayFactor, true, THXMLMessage.VALUE_GREATER_EQUAL_ZERO)
        specTable.storageDecayFactor = 1
    end
    local storageDryerXMLKey = xmlBaseKey .. ".storageDryer"
    local dryerTriggerNode = 0
    if siloSpec ~= nil and infoTriggerSpec ~= nil
        and infoTriggerSpec.infoTrigger ~= nil
        and self.getFillLevels ~= nil
    then
        local fillLevelList = self:getFillLevels()
        local hasMoistureFillType = false
        if fillLevelList ~= nil then
            for fillTypeIndex in pairs(fillLevelList) do
                if thMain:getFillTypeMoistureData(fillTypeIndex) ~= nil then
                    hasMoistureFillType = true
                    break
                end
            end
        end
        if hasMoistureFillType and infoTriggerSpec.infoTrigger > 0 then
            dryerTriggerNode = infoTriggerSpec.infoTrigger
        end
    end
    if xmlFile:hasProperty(storageDryerXMLKey) or dryerTriggerNode > 0 then
        local storageDryer = THStorageDryer.new(self)
        if storageDryer ~= nil and storageDryer:getIsEnabled() then
            if storageDryer:load(components, xmlFile, storageDryerXMLKey, customEnv, i3dMappings, dryerTriggerNode) then
                specTable.storageDryer = storageDryer
            else
                THUtils.stackMsg("Storage dryer failed to load")
                storageDryer:delete()
            end
        end
    end
end
function THPlaceableMoistureStorage:onFinalizePlacement()
    local specTable = getSpecTable(self)
    local ownerFarmId = self:getOwnerFarmId()
    local storageDryer = specTable.storageDryer
    if storageDryer ~= nil then
        storageDryer:register(true)
        storageDryer:setOwnerFarmId(ownerFarmId)
        storageDryer:setDryerStatus(THStorageDryer.DRYER_STATUS.STOPPED, true)
    end
end
function THPlaceableMoistureStorage:onDelete()
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        specTable.storageDryer:delete()
        specTable.storageDryer = nil
    end
end
function THPlaceableMoistureStorage:onReadStream(streamId, connection, ...)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        local storageDryerObjectId = NetworkUtil.readNodeObjectId(streamId)
        specTable.storageDryer:readStream(streamId, connection, ...)
        g_client:finishRegisterObject(specTable.storageDryer, storageDryerObjectId)
    end
    if connection:getIsServer() then
        self:thAA_readMoistureStorageFromStream(streamId, connection)
    end
end
function THPlaceableMoistureStorage:onWriteStream(streamId, connection, ...)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        local storageDryerObjectId = NetworkUtil.getObjectId(specTable.storageDryer)
        NetworkUtil.writeNodeObjectId(streamId, storageDryerObjectId)
        specTable.storageDryer:writeStream(streamId, connection, ...)
        g_server:registerObjectInStream(connection, specTable.storageDryer)
    end
    if not connection:getIsServer() then
        self:thAA_writeMoistureStorageToStream(streamId, connection)
    end
end
function THPlaceableMoistureStorage:onReadUpdateStream(streamId, timeStamp, connection, ...)
    if connection:getIsServer() then
        if streamReadBool(streamId) then -- isLossDirty
            self:thAA_readMoistureStorageFromStream(streamId, connection)
        end
    end
end
function THPlaceableMoistureStorage:onWriteUpdateStream(streamId, connection, dirtyMask, ...)
    local specTable = getSpecTable(self)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(specTable.dirtyFlag, dirtyMask) ~= 0) then
            self:thAA_writeMoistureStorageToStream(streamId, connection)
        end
    end
end
function THPlaceableMoistureStorage:onDraw()
    self:thAA_showMoistureHudBoxEntries()
end
function THPlaceableMoistureStorage:loadFromXMLFile(xmlFile, xmlKey)
end
function THPlaceableMoistureStorage:saveToXMLFile(xmlFile, xmlKey, usedModNames)
end
function THPlaceableMoistureStorage:thAA_readMoistureStorageFromStream(streamId, connection)
    local specTable = getSpecTable(self)
    THUtils.call(self.thAA_clearMoistureStorageLoss, self, nil, true)
    local numLossEntries = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
    if numLossEntries > 0 then
        for lossIndex = 1, numLossEntries do
            local fillTypeIndex = streamReadUIntN(streamId, FillTypeManager.SEND_NUM_BITS)
            local lossAmount = streamReadFloat32(streamId)
            local lossFactor = streamReadFloat32(streamId)
            local lossData = specTable.fillTypeLoss[fillTypeIndex]
            if lossData == nil then
                lossData = {
                    fillTypeIndex = fillTypeIndex,
                    amount = lossAmount,
                    factor = lossFactor
                }
                specTable.fillTypeLoss[fillTypeIndex] = lossData
            else
                lossData.amount = lossAmount
                lossData.factor = lossFactor
            end
            specTable.lossFillTypes[lossIndex] = fillTypeIndex
        end
    end
end
function THPlaceableMoistureStorage:thAA_writeMoistureStorageToStream(streamId, connection)
    local specTable = getSpecTable(self)
    local numLossEntries = #specTable.lossFillTypes
    streamWriteUIntN(streamId, numLossEntries, FillTypeManager.SEND_NUM_BITS)
    if numLossEntries > 0 then
        for lossIndex = 1, numLossEntries do
            local fillTypeIndex = specTable.lossFillTypes[lossIndex]
            local lossAmount = 0
            local lossFactor = 0
            local lossData = specTable.fillTypeLoss[fillTypeIndex]
            if lossData ~= nil then
                lossAmount = lossData.amount or 0
                lossFactor = lossData.factor or 0
            end
            streamWriteUIntN(streamId, fillTypeIndex, FillTypeManager.SEND_NUM_BITS)
            streamWriteFloat32(streamId, lossAmount)
            streamWriteFloat32(streamId, lossFactor)
        end
    end
end
function THPlaceableMoistureStorage:onMinuteChanged(currentMinute)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        specTable.storageDryer:updateDryer(1)
        if specTable.storageDryer:getDryerStatus() == THStorageDryer.DRYER_STATUS.RUNNING then
            specTable.dryerRunTime = specTable.dryerRunTime + 1
        end
    end
end
function THPlaceableMoistureStorage:onDayChanged(currentDay)
    local specTable = getSpecTable(self)
    if specTable.storageDryer ~= nil then
        specTable.storageDryer:processDryingCosts()
    end
    self:thAA_updateMoistureStorageLoss()
    specTable.dryerRunTime = 0
end
function THPlaceableMoistureStorage:thAA_clearMoistureStorageLoss(fillTypeIndex, noEventSend)
    local specTable = getSpecTable(self)
    if self.isServer or noEventSend == true then
        if fillTypeIndex ~= nil and fillTypeIndex ~= FillType.UNKNOWN then
            local lossData = specTable.fillTypeLoss[fillTypeIndex]
            if lossData ~= nil then
                lossData.amount = 0
                lossData.factor = 0
            end
        else
            THUtils.clearTable(specTable.lossFillTypes)
            for _, lossData in pairs(specTable.fillTypeLoss) do
                lossData.amount = 0
                lossData.factor = 0
            end
        end
        if self.isServer and noEventSend ~= true then
            self:raiseDirtyFlags(specTable.dirtyFlag)
        end
    end
end
function THPlaceableMoistureStorage:thAA_updateMoistureStorageLoss()
    local specTable = getSpecTable(self)
    local storageList, numStorages = self:thAA_getPlaceableObjects(THPlaceable.OBJECT_TYPE.STORAGE)
    if self.isServer then
        self:thAA_clearMoistureStorageLoss(nil, true)
        if storageList ~= nil and numStorages > 0 then
            local periodLength = thMain.periodLength or 1
            local globalDecayFactor = thSettingsManager:getSettingValue(THMoistureSystem.SETTING.STORAGE_LOSS_FACTOR) or 1
            local storageDecayFactor = specTable.storageDecayFactor
            if specTable.dryerRunTime > 0 then
                local dryerDecayFactor = thSettingsManager:getSettingValue(THMoistureSystem.SETTING.STORAGE_LOSS_FACTOR_DRYING) or 1
                local dryerRunPercent = MathUtil.clamp(specTable.dryerRunTime / 1440, 0, 1)
                globalDecayFactor = (dryerDecayFactor * dryerRunPercent) + (globalDecayFactor * (1 - dryerRunPercent))
            end
            storageDecayFactor = storageDecayFactor * globalDecayFactor
            for _, storage in pairs(storageList) do
                local thMoistureStorage = THMoistureStorage.getCustomData(storage)
                local storageFillLevels = storage:getFillLevels()
                if thMoistureStorage ~= nil and storageFillLevels ~= nil then
                    for fillTypeIndex, fillLevel in pairs(storageFillLevels) do
                        local fillTypeMoisture = thMoistureStorage:getFillTypeMoisture(fillTypeIndex)
                        if fillLevel > 0
                            and fillTypeMoisture ~= nil and fillTypeMoisture >= 0
                        then
                            local _, _, decayFactor = thMain:getFillTypeMoistureFactors(fillTypeIndex, fillTypeMoisture, false)
                            decayFactor = MathUtil.clamp((decayFactor or 0) * storageDecayFactor, 0, 1)
                            local decayAmount = (fillLevel * decayFactor) / periodLength
                            local newFillLevel = math.max(0, fillLevel - decayAmount)
                            if decayAmount >= 1 and newFillLevel ~= fillLevel then
                                storage:setFillLevel(newFillLevel, fillTypeIndex)
                                newFillLevel = storage:getFillLevel(fillTypeIndex)
                                decayAmount = math.max(0, fillLevel - newFillLevel)
                                if newFillLevel > 0 then
                                    decayFactor = decayAmount / newFillLevel
                                else
                                    decayFactor = 1
                                end
                            else
                                decayAmount = 0
                                decayFactor = 0
                            end
                            if decayAmount > 0 then
                                if THDebugUtil.getIsEnabled(debugFlagId) then
                                    local fillTypeName = thFillTypeManager:getFillTypeName(fillTypeIndex)
                                    THPlaceable.displayMsg(self, "Added storage loss for fillType %q [%s]: %s", fillTypeName, fillTypeIndex, decayAmount)
                                end
                            end
                            local lossData = specTable.fillTypeLoss[fillTypeIndex]
                            if lossData == nil then
                                lossData = {
                                    fillTypeIndex = fillTypeIndex,
                                    amount = decayAmount,
                                    factor = decayFactor
                                }
                                specTable.fillTypeLoss[fillTypeIndex] = lossData
                                table.insert(specTable.lossFillTypes, fillTypeIndex)
                            else
                                local totalLoss = lossData.amount + decayAmount
                                local lossFillPercent = decayAmount / totalLoss
                                lossData.amount = totalLoss
                                lossData.factor = (decayFactor * lossFillPercent) + (lossData.factor * (1 - lossFillPercent))
                            end
                        end
                    end
                end
            end
        end
        self:raiseDirtyFlags(specTable.dirtyFlag)
    end
end
function THPlaceableMoistureStorage:thAA_updateMoistureStorageBuffer()
    local specTable = getSpecTable(self)
    local moistureBuffer = THUtils.clearTable(specTable.moistureBuffer)
    local moistureBufferMapping = THUtils.clearTable(specTable.moistureBufferMapping)
    specTable.hasDisplayMoisture = false
    specTable.hasDisplayLoss = false
    local storageList, numStorages = self:thAA_getPlaceableObjects(THPlaceable.OBJECT_TYPE.STORAGE)
    if storageList ~= nil and numStorages > 0 then
        for storageIndex = 1, numStorages do
            local storage = storageList[storageIndex]
            local thMoistureStorage = THMoistureStorage.getCustomData(storage)
            local storageFillLevels = storage:getFillLevels()
            if thMoistureStorage ~= nil and storageFillLevels ~= nil then
                for fillTypeIndex, fillLevel in pairs(storageFillLevels) do
                    if fillLevel > 0 and fillTypeIndex ~= FillType.UNKNOWN then
                        local fillTypeMoistureData = thMain:getFillTypeMoistureData(fillTypeIndex)
                        local fillTypeMoisture = thMoistureStorage:getFillTypeMoisture(fillTypeIndex) or -1
                        if fillTypeMoistureData ~= nil then
                            local fillTypeInfo = fillTypeMoistureData.fillType.info
                            local bufferIndex = moistureBufferMapping[fillTypeIndex] or 0
                            local bufferData = moistureBuffer[bufferIndex]
                            if bufferIndex <= 0 or bufferData == nil then
                                bufferIndex = #moistureBuffer + 1
                                bufferData = {
                                    fillTypeIndex = fillTypeIndex,
                                    fillTypeTitle = fillTypeInfo.title,
                                    fillLevel = fillLevel,
                                    moisture = fillTypeMoisture,
                                    decayAmount = 0,
                                    decayFactor = 0
                                }
                                local lossData = specTable.fillTypeLoss[fillTypeIndex]
                                if lossData ~= nil then
                                    bufferData.decayAmount = lossData.amount or 0
                                    bufferData.decayFactor = lossData.factor or 0
                                end
                                moistureBuffer[bufferIndex] = bufferData
                                moistureBufferMapping[fillTypeIndex] = bufferIndex
                            else
                                local totalFillLevel = fillLevel or 0
                                local fillPercent = 1
                                if bufferData.fillLevel ~= nil and bufferData.fillLevel > 0 then
                                    totalFillLevel = bufferData.fillLevel + fillLevel
                                    fillPercent = fillLevel / totalFillLevel
                                end
                                if totalFillLevel > 0 and fillTypeMoisture >= 0 then
                                    local minMoisture = fillTypeMoistureData.minPercent
                                    local maxMoisture = fillTypeMoistureData.maxPercent
                                    if bufferData.moisture == nil or bufferData.moisture < 0 then
                                        bufferData.moisture = fillTypeMoisture
                                    else
                                        bufferData.moisture = (fillTypeMoisture * fillPercent) + (bufferData.moisture * (1 - fillPercent))
                                        bufferData.moisture = MathUtil.clamp(bufferData.moisture, minMoisture, maxMoisture)
                                    end
                                else
                                    bufferData.moisture = -1
                                end
                                bufferData.fillLevel = totalFillLevel
                            end
                            if bufferData.moisture ~= nil and bufferData.moisture >= 0 then
                                specTable.hasDisplayMoisture = true
                            end
                            if bufferData.decayAmount ~= nil and bufferData.decayAmount > 0 then
                                specTable.hasDisplayLoss = true
                            end
                        end
                    end
                end
            end
        end
    end
    local numBufferEntries = #moistureBuffer
    if numBufferEntries > 0 then
        table.sort(moistureBuffer, function(arg1, arg2)
            return arg1.fillLevel > arg2.fillLevel
        end)
    end
    return moistureBuffer, numBufferEntries
end
function THPlaceableMoistureStorage:thAA_showMoistureHudBoxEntries()
    local specTable = getSpecTable(self)
    local mission = thMain.mission
    local ownerFarmId = self:getOwnerFarmId()
    local moistureHudBox = specTable.moistureHudBox
    moistureHudBox:clear()
    if mission ~= nil and specTable.isMoistureHudBoxEnabled
        and (ownerFarmId == AccessHandler.EVERYONE or ownerFarmId == mission:getFarmId())
    then
        local moistureBuffer, numBufferEntries = self:thAA_updateMoistureStorageBuffer()
        if moistureBuffer ~= nil and numBufferEntries > 0
            and (specTable.hasDisplayMoisture or specTable.hasDisplayLoss)
        then
            local siloName = self:getName() or specTable.hudTitles.storage
            moistureHudBox:setTitle(siloName)
            if specTable.hasDisplayMoisture then
                moistureHudBox:addLine(specTable.hudTitles.moisture, "", true)
                for bufferIndex = 1, numBufferEntries do
                    local bufferData = moistureBuffer[bufferIndex]
                    if bufferData.moisture ~= nil and bufferData.moisture >= 0 then
                        local moistureText = THUtils.formatPercent(bufferData.moisture, 1)
                        moistureHudBox:addLine("- " .. bufferData.fillTypeTitle, moistureText, false)
                    end
                end
            end
            if specTable.hasDisplayLoss then
                moistureHudBox:addLine(specTable.hudTitles.storageLoss, "", true)
                for bufferIndex = 1, numBufferEntries do
                    local bufferData = moistureBuffer[bufferIndex]
                    local lossText = thI18n:formatVolume(bufferData.decayAmount or 0)
                    local lossPercentText = THUtils.formatPercent(bufferData.decayFactor or 0, 1)
                    lossText = string.format("%s (%s)", lossText, lossPercentText)
                    moistureHudBox:addLine("- " .. bufferData.fillTypeTitle, lossText, false)
                end
            end
            if specTable.storageDryer ~= nil then
                local dryerStatus = specTable.storageDryer:getDryerStatus()
                if dryerStatus ~= THStorageDryer.DRYER_STATUS.DISABLED then
                    local dryerStatusText = specTable.storageDryer:getDryerStatusText()
                    if dryerStatusText ~= nil then
                        local statusTitle = specTable.hudTitles.dryerStatus
                        moistureHudBox:addLine(statusTitle, dryerStatusText)
                    end
                end
            end
            moistureHudBox:showNextFrame()
        end
    end
end
function THPlaceableMoistureStorage:setOwnerFarmId(superFunc, farmId, noEventSend, ...)
    local specTable = getSpecTable(self)
    local function appendFunc(...)
        if specTable.storageDryer ~= nil then
            THUtils.call(function()
                specTable.storageDryer:setOwnerFarmId(farmId, noEventSend)
            end)
        end
        return ...
    end
    return appendFunc(superFunc(self, farmId, noEventSend, ...))
end
function THPlaceableMoistureStorage:collectPickObjects(superFunc, nodeId, ...)
    local specTable = getSpecTable(self)
    local isNodeIdFound = false
    if specTable.storageDryer ~= nil then
        local dryerData = THStorageDryer.getCustomData(specTable.storageDryer)
        if dryerData ~= nil and nodeId == dryerData.interactionTrigger then
            isNodeIdFound = true
        end
    end
    if not isNodeIdFound then
        return superFunc(self, nodeId, ...)
    end
end
function THPlaceableMoistureStorage:getNeedMinuteChanged(superFunc, ...)
    local function appendFunc(rNeedChange, ...)
        rNeedChange = true
        return rNeedChange, ...
    end
    return appendFunc(superFunc(self, ...))
end
function THPlaceableMoistureStorage:getNeedDayChanged(superFunc, ...)
    local function appendFunc(rNeedChange, ...)
        rNeedChange = true
        return rNeedChange, ...
    end
    return appendFunc(superFunc(self, ...))
end